from __future__ import annotations

import datetime
import types
from abc import abstractmethod
from collections import UserList
from collections.abc import Collection, Mapping
from typing import Any, ClassVar, Generic, Optional, TypeVar, Union

import pandas as pd
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import Properties, PropertyValue
from pydantic import BaseModel, ConfigDict, Extra, Field


class DomainModelCore(BaseModel):
    space: ClassVar[str]
    external_id: str = Field(min_length=1, max_length=255)

    def id_tuple(self) -> tuple[str, str]:
        return self.space, self.external_id


class DomainModel(DomainModelCore):
    version: int
    last_updated_time: datetime.datetime
    created_time: datetime.datetime
    deleted_time: Optional[datetime.datetime] = None

    @classmethod
    def from_node(cls: type[T_TypeNode], node: dm.Node) -> T_TypeNode:
        data = node.dump(camel_case=False)

        return cls(**{**data, **unpack_properties(node.properties)})  # type: ignore[arg-type]

    @classmethod
    def one_to_many_fields(cls) -> list[str]:
        return [
            field_name
            for field_name, field in cls.model_fields.items()
            if isinstance(field.annotation, types.GenericAlias)
        ]


T_TypeNode = TypeVar("T_TypeNode", bound=DomainModel)


class DomainModelApply(DomainModelCore):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra=Extra.forbid)
    existing_version: Optional[int] = None

    def to_instances_apply(self) -> dm.InstancesApply:
        return self._to_instances_apply(set())

    @abstractmethod
    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        raise NotImplementedError()


class DomainModelApplyResult(DomainModelCore):
    version: int
    was_modified: bool
    last_updated_time: datetime.datetime
    created_time: datetime.datetime


class DataPoint(BaseModel):
    timestamp: str


class NumericDataPoint(DataPoint):
    value: float


class StringDataPoint(DataPoint):
    value: str


class TimeSeries(DomainModelCore):
    id: Optional[int] = None
    name: Optional[str] = None
    is_string: bool = False
    metadata: dict[str, str] = Field(default_factory=dict)
    unit: Optional[str] = None
    asset_id: Optional[int] = None
    is_step: bool = False
    description: Optional[str] = None
    security_categories: Optional[str] = None
    dataset_id: Optional[int] = None
    data_points: Union[list[NumericDataPoint], list[StringDataPoint]]


class TypeList(UserList, Generic[T_TypeNode]):
    _NODE: type[T_TypeNode]

    def __init__(self, nodes: Collection[type[DomainModelCore]]):
        super().__init__(nodes)

    def dump(self) -> list[dict[str, Any]]:
        return [node.dict() for node in self.data]

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame(self.dump())

    def _repr_html_(self) -> str:
        return self.to_pandas()._repr_html_()  # type: ignore[operator]


T_TypeApplyNode = TypeVar("T_TypeApplyNode", bound=DomainModelApply)
T_TypeNodeList = TypeVar("T_TypeNodeList", bound=TypeList)


def unpack_properties(properties: Properties) -> Mapping[str, PropertyValue]:
    unpacked: dict[str, PropertyValue] = {}
    for view_properties in properties.values():
        for prop_name, prop_value in view_properties.items():
            if isinstance(prop_value, (str, int, float, bool, list)):
                unpacked[prop_name] = prop_value
            elif isinstance(prop_value, dict):
                # Dicts are assumed to be reference properties
                if "space" in prop_value and "externalId" in prop_value:
                    unpacked[prop_name] = prop_value["externalId"]
                else:
                    raise ValueError(f"Unexpected reference property {prop_value}")
            else:
                raise ValueError(f"Unexpected property value type {type(prop_value)}")
    return unpacked



from __future__ import annotations

import json
from enum import Enum, auto
from typing import Any, ClassVar, Iterator, Optional
from typing_extensions import TypeAlias, Annotated
import pandas as pd
from pydantic import BaseModel, constr, validator

ExternalId: TypeAlias = Annotated[str, constr(min_length=1, max_length=255)]


class Auction(str, Enum):
    week = "week"
    weekend = "weekend"


class RetrievalType(Enum):
    RANGE = auto()
    START = auto()
    END = auto()


class AggregationMethod(Enum):
    sum = auto()
    mean = auto()
    std = auto()
    sem = auto()
    max = auto()
    min = auto()
    median = auto()
    first = auto()
    last = auto()


class TransformationType(Enum):
    GATE_SCHEDULE_FLAG_VALUE_MAPPING = auto()
    GENERATOR_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING = auto()
    PLANT_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING = auto()
    GATE_OPENING_METER_TO_PERCENT = auto()
    TO_BOOL = auto()
    ZERO_IF_NOT_ONE = auto()
    MULTIPLY = auto()
    MULTIPLY_FROM_OFFSET = auto()
    ADD = auto()
    ADD_FROM_OFFSET = auto()
    DYNAMIC_ADD_FROM_OFFSET = auto()
    RESERVOIR_LEVEL_TO_VOLUME = auto()
    STATIC = auto()
    DYNAMIC_STATIC = auto()
    ONE_IF_TWO = auto()
    ADD_WATER_IN_TRANSIT = auto()


ATTRIBUTE_DEFAULT_AGGREGATION: dict[str, AggregationMethod | None] = {
    "gate.max_flow": AggregationMethod.min,
    "gate.min_flow": AggregationMethod.max,
    "gate.schedule_flag": AggregationMethod.first,
    "gate.schedule_m3s": AggregationMethod.first,
    "generator.committed_flag": AggregationMethod.first,
    "generator.committed_in": None,
    "generator.initial_state": None,
    "generator.maintenance_flag": AggregationMethod.max,
    "generator.max_p_constr": AggregationMethod.min,
    "generator.min_p_constr": AggregationMethod.max,
    "generator.production_schedule_flag": AggregationMethod.first,
    "market.buy_price": AggregationMethod.mean,
    "market.sale_price": AggregationMethod.mean,
    "plant.feeding_fee": AggregationMethod.mean,
    "plant.maintenance_flag": AggregationMethod.max,
    "plant.max_p_constr": AggregationMethod.min,
    "plant.max_q_constr": AggregationMethod.min,
    "plant.min_p_constr": AggregationMethod.max,
    "plant.mip_flag": AggregationMethod.first,
    "plant.production_schedule": AggregationMethod.first,
    "plant.production_schedule_flag": AggregationMethod.first,
    "reserve_group.rr_up_obligation": AggregationMethod.max,
    "reserve_group.rr_down_obligation": AggregationMethod.max,
    "reservoir.inflow": AggregationMethod.mean,
    "reservoir.max_vol_constr": AggregationMethod.min,
    "reservoir.min_vol_constr": AggregationMethod.max,
    "reservoir.start_head": None,
    "reservoir.tactical_cost_max": AggregationMethod.mean,
    "reservoir.tactical_cost_min": AggregationMethod.mean,
    "reservoir.tactical_limit_max": AggregationMethod.mean,
    "reservoir.tactical_limit_min": AggregationMethod.mean,
}


class Transformation(BaseModel):  # type: ignore[no-redef]  # mypy says kwargs is redefined on this line ¯\_(ツ)_/¯
    transformation: TransformationType
    kwargs: Optional[dict] = None

    @validator("transformation", pre=True)
    def to_type(cls, value):
        return TransformationType[value.upper()] if isinstance(value, str) else value

    def to_dict(self) -> dict:
        return {"transformation": self.transformation.name, "kwargs": self.kwargs}


class TimeSeriesMappingEntry(BaseModel):
    object_type: str
    object_name: str
    attribute_name: str
    time_series_external_id: Optional[str] = None
    transformations: Optional[list[Transformation]] = None
    retrieve: Optional[RetrievalType] = None
    aggregation: Optional[AggregationMethod] = None

    @validator("aggregation", pre=True)
    def to_enum(cls, value):
        return AggregationMethod[value] if isinstance(value, str) else value

    @validator("aggregation", always=True)
    def set_default(cls, value, values):
        if value is not None:
            return value
        # TODO: do we want default `None` here or raise error?
        return ATTRIBUTE_DEFAULT_AGGREGATION.get(f"{values.get('object_type')}.{values.get('attribute_name')}")

    @validator("retrieve", pre=True)
    def to_retrival_enum(cls, value):
        return RetrievalType[value] if isinstance(value, str) else value

    @property
    def shop_model_path(self) -> str:
        return f"{self.object_type}.{self.object_name}.{self.attribute_name}"

    def _transformations_to_strings(self, max_cols: int) -> list[str]:
        if self.transformations:
            # ensure_ascii=False to allow æøåÆØÅ
            transformation_string = json.dumps([t.to_dict() for t in self.transformations], ensure_ascii=False)
        else:
            transformation_string = json.dumps([])

        max_chars = 255
        return [transformation_string[i * max_chars : (i + 1) * max_chars] for i in range(max_cols)]

    def to_sequence_row(self, max_transformation_cols: int) -> list[str | float]:
        return [
            self.shop_model_path,
            self.time_series_external_id or float("nan"),
            *self._transformations_to_strings(max_cols=max_transformation_cols),
            self.retrieve.name if self.retrieve else float("nan"),
            self.aggregation.name if self.aggregation else float("nan"),
        ]


class TimeSeriesMapping(BaseModel):
    rows: list[TimeSeriesMappingEntry] = []
    columns: ClassVar[list[str]] = [
        "shop_model_path",
        "time_series_external_id",
        "transformations",
        "transformations1",
        "transformations2",
        "transformations3",
        "retrieve",
        "aggregation",
    ]

    @property
    def transformations_cols(self) -> list[str]:
        return [col for col in self.columns if col.startswith("transformations")]

    def __iter__(self) -> Iterator[TimeSeriesMappingEntry]:  # type: ignore[override]
        yield from self.rows

    def __len__(self) -> int:
        return len(self.rows)

    def __add__(self, other: "TimeSeriesMapping") -> "TimeSeriesMapping":
        return TimeSeriesMapping(rows=self.rows + other.rows)

    def append(self, element: TimeSeriesMappingEntry) -> None:
        self.rows.append(element)

    def extend(self, other: "TimeSeriesMapping") -> None:
        self.rows.extend(other.rows)

    @property
    def column_definitions(self) -> list[dict]:
        return [{"valueType": "STRING", "externalId": col} for col in self.columns]

    def to_dataframe(self) -> pd.DataFrame:
        rows = [row.to_sequence_row(max_transformation_cols=len(self.transformations_cols)) for row in self.rows]
        return pd.DataFrame(data=rows, columns=self.columns)

    def dumps(self) -> dict[str, Any]:
        rows = []
        for row in self.rows:
            row_raw = row.dict(exclude={"aggregation", "retrieve", "transformations"})
            if row.aggregation:
                row_raw["aggregation"] = row.aggregation.name
            if row.retrieve:
                row_raw["retrieve"] = row.retrieve.name
            if row.transformations:
                row_raw["transformations"] = [
                    {
                        **transformation.dict(exclude={"transformation"}),
                        "transformation": transformation.transformation.name,
                    }
                    for transformation in row.transformations
                ]
            rows.append(row_raw)
        return {"rows": rows}



from cognite.powerops.clients.data_classes._benchmark_bids import BenchmarkBid, BenchmarkBidApply, BenchmarkBidList
from cognite.powerops.clients.data_classes._benchmark_process import (
    BenchmarkProces,
    BenchmarkProcesApply,
    BenchmarkProcesList,
)
from cognite.powerops.clients.data_classes._bid_matrix_generators import (
    BidMatrixGenerator,
    BidMatrixGeneratorApply,
    BidMatrixGeneratorList,
)
from cognite.powerops.clients.data_classes._command_configs import CommandConfig, CommandConfigApply, CommandConfigList
from cognite.powerops.clients.data_classes._date_transformations import (
    DateTransformation,
    DateTransformationApply,
    DateTransformationList,
)
from cognite.powerops.clients.data_classes._day_ahead_bids import DayAheadBid, DayAheadBidApply, DayAheadBidList
from cognite.powerops.clients.data_classes._day_ahead_process import (
    DayAheadProces,
    DayAheadProcesApply,
    DayAheadProcesList,
)
from cognite.powerops.clients.data_classes._generators import Generator, GeneratorApply, GeneratorList
from cognite.powerops.clients.data_classes._input_time_series_mappings import (
    InputTimeSeriesMapping,
    InputTimeSeriesMappingApply,
    InputTimeSeriesMappingList,
)
from cognite.powerops.clients.data_classes._nord_pool_markets import (
    NordPoolMarket,
    NordPoolMarketApply,
    NordPoolMarketList,
)
from cognite.powerops.clients.data_classes._output_containers import (
    OutputContainer,
    OutputContainerApply,
    OutputContainerList,
)
from cognite.powerops.clients.data_classes._output_mappings import OutputMapping, OutputMappingApply, OutputMappingList
from cognite.powerops.clients.data_classes._plants import Plant, PlantApply, PlantList
from cognite.powerops.clients.data_classes._price_areas import PriceArea, PriceAreaApply, PriceAreaList
from cognite.powerops.clients.data_classes._production_plan_time_series import (
    ProductionPlanTimeSeries,
    ProductionPlanTimeSeriesApply,
    ProductionPlanTimeSeriesList,
)
from cognite.powerops.clients.data_classes._rkom_bids import RKOMBid, RKOMBidApply, RKOMBidList
from cognite.powerops.clients.data_classes._rkom_bid_combinations import (
    RKOMBidCombination,
    RKOMBidCombinationApply,
    RKOMBidCombinationList,
)
from cognite.powerops.clients.data_classes._rkom_combination_bids import (
    RKOMCombinationBid,
    RKOMCombinationBidApply,
    RKOMCombinationBidList,
)
from cognite.powerops.clients.data_classes._rkom_markets import RKOMMarket, RKOMMarketApply, RKOMMarketList
from cognite.powerops.clients.data_classes._rkom_process import RKOMProces, RKOMProcesApply, RKOMProcesList
from cognite.powerops.clients.data_classes._reserve_scenarios import (
    ReserveScenario,
    ReserveScenarioApply,
    ReserveScenarioList,
)
from cognite.powerops.clients.data_classes._reservoirs import Reservoir, ReservoirApply, ReservoirList
from cognite.powerops.clients.data_classes._scenarios import Scenario, ScenarioApply, ScenarioList
from cognite.powerops.clients.data_classes._scenario_mappings import (
    ScenarioMapping,
    ScenarioMappingApply,
    ScenarioMappingList,
)
from cognite.powerops.clients.data_classes._scenario_templates import (
    ScenarioTemplate,
    ScenarioTemplateApply,
    ScenarioTemplateList,
)
from cognite.powerops.clients.data_classes._shop_transformations import (
    ShopTransformation,
    ShopTransformationApply,
    ShopTransformationList,
)
from cognite.powerops.clients.data_classes._value_transformations import (
    ValueTransformation,
    ValueTransformationApply,
    ValueTransformationList,
)
from cognite.powerops.clients.data_classes._watercourses import Watercourse, WatercourseApply, WatercourseList
from cognite.powerops.clients.data_classes._watercourse_shops import (
    WatercourseShop,
    WatercourseShopApply,
    WatercourseShopList,
)

BenchmarkBidApply.model_rebuild()
BenchmarkProcesApply.model_rebuild()
DayAheadBidApply.model_rebuild()
DayAheadProcesApply.model_rebuild()
InputTimeSeriesMappingApply.model_rebuild()
OutputContainerApply.model_rebuild()
PlantApply.model_rebuild()
PriceAreaApply.model_rebuild()
RKOMBidApply.model_rebuild()
RKOMBidCombinationApply.model_rebuild()
RKOMProcesApply.model_rebuild()
ScenarioApply.model_rebuild()
ScenarioMappingApply.model_rebuild()
ScenarioTemplateApply.model_rebuild()
ShopTransformationApply.model_rebuild()
WatercourseApply.model_rebuild()

__all__ = [
    "BenchmarkBid",
    "BenchmarkBidApply",
    "BenchmarkBidList",
    "BenchmarkProces",
    "BenchmarkProcesApply",
    "BenchmarkProcesList",
    "BidMatrixGenerator",
    "BidMatrixGeneratorApply",
    "BidMatrixGeneratorList",
    "CommandConfig",
    "CommandConfigApply",
    "CommandConfigList",
    "DateTransformation",
    "DateTransformationApply",
    "DateTransformationList",
    "DayAheadBid",
    "DayAheadBidApply",
    "DayAheadBidList",
    "DayAheadProces",
    "DayAheadProcesApply",
    "DayAheadProcesList",
    "Generator",
    "GeneratorApply",
    "GeneratorList",
    "InputTimeSeriesMapping",
    "InputTimeSeriesMappingApply",
    "InputTimeSeriesMappingList",
    "NordPoolMarket",
    "NordPoolMarketApply",
    "NordPoolMarketList",
    "OutputContainer",
    "OutputContainerApply",
    "OutputContainerList",
    "OutputMapping",
    "OutputMappingApply",
    "OutputMappingList",
    "Plant",
    "PlantApply",
    "PlantList",
    "PriceArea",
    "PriceAreaApply",
    "PriceAreaList",
    "ProductionPlanTimeSeries",
    "ProductionPlanTimeSeriesApply",
    "ProductionPlanTimeSeriesList",
    "RKOMBid",
    "RKOMBidApply",
    "RKOMBidList",
    "RKOMBidCombination",
    "RKOMBidCombinationApply",
    "RKOMBidCombinationList",
    "RKOMCombinationBid",
    "RKOMCombinationBidApply",
    "RKOMCombinationBidList",
    "RKOMMarket",
    "RKOMMarketApply",
    "RKOMMarketList",
    "RKOMProces",
    "RKOMProcesApply",
    "RKOMProcesList",
    "ReserveScenario",
    "ReserveScenarioApply",
    "ReserveScenarioList",
    "Reservoir",
    "ReservoirApply",
    "ReservoirList",
    "Scenario",
    "ScenarioApply",
    "ScenarioList",
    "ScenarioMapping",
    "ScenarioMappingApply",
    "ScenarioMappingList",
    "ScenarioTemplate",
    "ScenarioTemplateApply",
    "ScenarioTemplateList",
    "ShopTransformation",
    "ShopTransformationApply",
    "ShopTransformationList",
    "ValueTransformation",
    "ValueTransformationApply",
    "ValueTransformationList",
    "Watercourse",
    "WatercourseApply",
    "WatercourseList",
    "WatercourseShop",
    "WatercourseShopApply",
    "WatercourseShopList",
]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.clients.data_classes._benchmark_bids import BenchmarkBidApply
    from cognite.powerops.clients.data_classes._production_plan_time_series import ProductionPlanTimeSeriesApply
    from cognite.powerops.clients.data_classes._shop_transformations import ShopTransformationApply

__all__ = ["BenchmarkProces", "BenchmarkProcesApply", "BenchmarkProcesList"]


class BenchmarkProces(DomainModel):
    space: ClassVar[str] = "power-ops"
    bid: Optional[str] = None
    metrics: Optional[dict] = None
    name: Optional[str] = None
    production_plan_time_series: list[str] = Field([], alias="productionPlanTimeSeries")
    run_events: list[str] = Field([], alias="runEvents")
    shop: Optional[str] = None


class BenchmarkProcesApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    bid: Optional[Union["BenchmarkBidApply", str]] = Field(None, repr=False)
    metrics: Optional[dict] = None
    name: Optional[str] = None
    production_plan_time_series: list[Union["ProductionPlanTimeSeriesApply", str]] = Field(
        default_factory=list, repr=False
    )
    run_events: list[str] = []
    shop: Optional[Union["ShopTransformationApply", str]] = Field(None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "BenchmarkProcess"),
            properties={
                "bid": {
                    "space": "power-ops",
                    "externalId": self.bid if isinstance(self.bid, str) else self.bid.external_id,
                },
                "metrics": self.metrics,
                "name": self.name,
                "runEvents": self.run_events,
                "shop": {
                    "space": "power-ops",
                    "externalId": self.shop if isinstance(self.shop, str) else self.shop.external_id,
                },
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        for production_plan_time_series in self.production_plan_time_series:
            edge = self._create_production_plan_time_series_edge(production_plan_time_series)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(production_plan_time_series, DomainModelApply):
                instances = production_plan_time_series._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.bid, DomainModelApply):
            instances = self.bid._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.shop, DomainModelApply):
            instances = self.shop._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_production_plan_time_series_edge(
        self, production_plan_time_series: Union[str, "ProductionPlanTimeSeriesApply"]
    ) -> dm.EdgeApply:
        if isinstance(production_plan_time_series, str):
            end_node_ext_id = production_plan_time_series
        elif isinstance(production_plan_time_series, DomainModelApply):
            end_node_ext_id = production_plan_time_series.external_id
        else:
            raise TypeError(f"Expected str or ProductionPlanTimeSeriesApply, got {type(production_plan_time_series)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "BenchmarkProcess.productionPlanTimeSeries"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class BenchmarkProcesList(TypeList[BenchmarkProces]):
    _NODE = BenchmarkProces



from __future__ import annotations

from pathlib import Path

from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import OAuthClientCredentials

from cognite.powerops.cogshop1._api.cases import CasesAPI
from cognite.powerops.cogshop1._api.commands_configs import CommandsConfigsAPI
from cognite.powerops.cogshop1._api.file_refs import FileRefsAPI
from cognite.powerops.cogshop1._api.mappings import MappingsAPI
from cognite.powerops.cogshop1._api.model_templates import ModelTemplatesAPI
from cognite.powerops.cogshop1._api.processing_logs import ProcessingLogsAPI
from cognite.powerops.cogshop1._api.scenarios import ScenariosAPI
from cognite.powerops.cogshop1._api.transformations import TransformationsAPI


class CogShop1Client:
    """
    CogShop1Client

    Generated with:
        pygen = 0.15.3
        cognite-sdk = 6.15.0
        pydantic = 2.3.0

    Data Model:
        space: cogShop
        externalId: CogShop
        version: 2
    """

    def __init__(self, config_or_client: CogniteClient | ClientConfig):
        if isinstance(config_or_client, CogniteClient):
            client = config_or_client
        elif isinstance(config_or_client, ClientConfig):
            client = CogniteClient(config_or_client)
        else:
            raise ValueError(f"Expected CogniteClient or ClientConfig, got {type(config_or_client)}")
        self.cases = CasesAPI(client)
        self.commands_configs = CommandsConfigsAPI(client)
        self.file_refs = FileRefsAPI(client)
        self.mappings = MappingsAPI(client)
        self.model_templates = ModelTemplatesAPI(client)
        self.processing_logs = ProcessingLogsAPI(client)
        self.scenarios = ScenariosAPI(client)
        self.transformations = TransformationsAPI(client)

    @classmethod
    def azure_project(
        cls, tenant_id: str, client_id: str, client_secret: str, cdf_cluster: str, project: str
    ) -> CogShop1Client:
        credentials = OAuthClientCredentials.default_for_azure_ad(tenant_id, client_id, client_secret, cdf_cluster)
        config = ClientConfig.default(project, cdf_cluster, credentials)

        return cls(config)

    @classmethod
    def from_toml(cls, file_path: Path | str, section: str | None = "cognite") -> CogShop1Client:
        import toml

        toml_content = toml.load(file_path)
        if section is not None:
            try:
                toml_content = toml_content[section]
            except KeyError:
                raise ValueError(f"Could not find section '{section}' in {file_path}")

        return cls.azure_project(**toml_content)



from __future__ import annotations

from typing import ClassVar, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

__all__ = ["WatercourseShop", "WatercourseShopApply", "WatercourseShopList"]


class WatercourseShop(DomainModel):
    space: ClassVar[str] = "power-ops"
    penalty_limit: Optional[float] = Field(None, alias="penaltyLimit")


class WatercourseShopApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    penalty_limit: Optional[float] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "WatercourseShop"),
            properties={
                "penaltyLimit": self.penalty_limit,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class WatercourseShopList(TypeList[WatercourseShop]):
    _NODE = WatercourseShop



from __future__ import annotations

import json
from abc import ABC
from collections import defaultdict, Counter
from typing import ClassVar, Union, Optional, TypeVar
from typing_extensions import Self

from cognite.client.data_classes import Relationship, TimeSeries, Label, Asset
from pydantic import BaseModel

from cognite.powerops.cdf_labels import AssetLabel, RelationshipLabel
from .resource_type import ResourceType
from cognite.powerops.resync.models.cdf_resources import CDFSequence, CDFFile
from cognite.powerops.resync.utils.serializer import get_pydantic_annotation


class NonAssetType(BaseModel, ABC, arbitrary_types_allowed=True):
    ...


class AssetType(ResourceType, ABC, arbitrary_types_allowed=True, validate_assignment=True):
    parent_external_id: ClassVar[str]
    label: ClassVar[Union[AssetLabel, str]]
    parent_description: ClassVar[Optional[str]] = None
    _instantiated_assets: ClassVar[dict[str, AssetType]] = defaultdict(dict)

    name: str
    description: Optional[str] = None
    _external_id: Optional[str] = None
    _type: Optional[str] = None

    @property
    def external_id(self) -> str:
        return self._external_id or f"{self.type_}_{self.name}"

    @external_id.setter
    def external_id(self, value: str) -> None:
        """
        This setter is only used in the Market models which have inconsistent naming between the type and the
        parent_external_id. It is a workaround to keep the refactoring to models to avoid introducing breaking changes
        in CDF.
        Parameters
        ----------
        value : str
            The external id of the asset.

        """
        self._external_id = value

    @property
    def type_(self) -> str:
        return self._type or self.parent_external_id.removesuffix("s")

    @type_.setter
    def type_(self, value: str) -> None:
        """
        This setter is only used in the Market models which have inconsistent naming between the type and
        parent_external_id. It is a workaround to keep the refactoring to models to avoid introducing breaking changes
        in CDF.

        Parameters
        ----------
        value: str
            The type of the asset.

        """
        self._type = value

    @property
    def parent_name(self):
        return self.parent_external_id.replace("_", " ").title()

    def relationships(self) -> list[Relationship]:
        return self._relationships(self)

    def _relationships(self, to: BaseModel) -> list[Relationship]:
        relationships = []
        for field_name, field in to.model_fields.items():
            field_value = getattr(to, field_name)
            if not field_value:
                continue
            if isinstance(field_value, CDFSequence):
                relationships.append(self._create_relationship(field_value.external_id, "SEQUENCE", field_name))
            elif isinstance(field_value, TimeSeries):
                relationships.append(self._create_relationship(field_value.external_id, "TIMESERIES", field_name))
            elif isinstance(field_value, list) and field_value and isinstance(field_value[0], TimeSeries):
                # The string comparison is to avoid circular imports
                if type(to).__name__ == "ProductionPlanTimeSeries" and field_name == "series":
                    # For some reason, the production plan time series are not linked to the benchmarking config
                    continue

                relationships.extend(
                    self._create_relationship(ts.external_id, "TIMESERIES", field_name) for ts in field_value
                )
            elif isinstance(field_value, list) and field_value and isinstance(field_value[0], CDFSequence):
                relationships.extend(
                    self._create_relationship(sequence.external_id, "SEQUENCE", field_name) for sequence in field_value
                )
            elif isinstance(field_value, AssetType) or (
                isinstance(field_value, list) and field_value and isinstance(field_value[0], AssetType)
            ):
                asset_types: list[AssetType] = field_value if isinstance(field_value, list) else [field_value]
                for target in asset_types:
                    target_type = target.type_
                    if self.type_ == "plant" and target.type_ == "reservoir":
                        target_type = "inlet_reservoir"
                    relationships.append(self._create_relationship(target.external_id, "ASSET", target_type))
            elif isinstance(field_value, NonAssetType) or (
                isinstance(field_value, list) and field_value and isinstance(field_value[0], NonAssetType)
            ):
                non_asset_types: list[NonAssetType] = field_value if isinstance(field_value, list) else [field_value]
                for target in non_asset_types:
                    relationships.extend(self._relationships(target))

        return relationships

    def _create_relationship(
        self,
        target_external_id: str,
        target_cdf_type: str,
        target_type: str,
    ) -> Relationship:
        source_type = "ASSET"

        # The market model uses the suffix CDF type for the relationship label, while the core model does not.
        try:
            # Production Model
            label = RelationshipLabel(f"relationship_to.{target_type}")
        except ValueError:
            # Market Model
            label = RelationshipLabel(f"relationship_to.{target_type}_{target_cdf_type.lower()}")
            # In addition, the Market Model uses capitalised CDF types for the relationship type,
            # while the core model uses all upper cases.
            source_type = source_type.title()
            target_cdf_type = target_cdf_type.title()

        return Relationship(
            external_id=f"{self.external_id}.{target_external_id}",
            source_external_id=self.external_id,
            source_type=source_type,
            target_external_id=target_external_id,
            target_type=target_cdf_type,
            labels=[Label(external_id=label.value)],
        )

    def as_asset(self) -> Asset:
        metadata = self._as_metadata()

        return Asset(
            external_id=self.external_id,
            name=self.name,
            parent_external_id=self.parent_external_id,
            labels=[Label(str(self.label.value))],
            metadata=metadata or None,
            description=self.description,
        )

    def _as_metadata(self) -> dict[str, str]:
        metadata = {}
        for field_name, field in self.model_fields.items():
            annotation, _ = get_pydantic_annotation(field.annotation, type(self))
            if (
                annotation in [CDFSequence, TimeSeries, CDFFile]
                or issubclass(annotation, AssetType)
                or field_name in {"name", "description", "label", "parent_external_id"}
                or field.exclude
            ):
                continue

            value = getattr(self, field_name)
            if isinstance(value, NonAssetType):
                value = value.model_dump(exclude_unset=True)
                for k, v in value.items():
                    if isinstance(v, (dict, list)):
                        v = json.dumps(v)
                    metadata[f"{field_name}:{k}"] = v
            elif isinstance(value, list) and value and isinstance(value[0], NonAssetType):
                metadata[field_name] = json.dumps([item.model_dump(exclude_unset=True) for item in value])
            elif isinstance(value, (dict, list)) and value:
                metadata[field_name] = json.dumps(value)
            elif isinstance(value, (dict, list)) and not value:
                continue
            elif field.annotation in (str, int, float):
                metadata[field_name] = value
            else:
                raise NotImplementedError(f"Cannot handle metadata of type {field.annotation}")
        return metadata

    @classmethod
    def from_asset(cls, asset: Asset) -> Self:
        instance = cls(
            name=asset.name,
            description=asset.description,
            **cls._load_metadata(asset.metadata),
        )
        if instance.external_id != asset.external_id:
            # The external id is not set in a standardized way for Market configs.
            instance.external_id = asset.external_id
        return instance

    @classmethod
    def _load_metadata(cls, metadata: dict[str, str] | None = None) -> dict[str, dict | str]:
        if not metadata:
            return {}
        output: dict[str, dict | str] = {}
        for key, value in metadata.items():
            counts = Counter(key)
            if counts[":"] == 1:
                field_name, sub_key = key.split(":")
                if field_name not in output:
                    output[field_name] = {}
                output[field_name][sub_key] = value
            elif counts[":"] == 0:
                output[key] = value
            else:
                raise NotImplementedError("Nested keys more than one level is not supported")

        return output


T_Asset_Type = TypeVar("T_Asset_Type", bound=AssetType)



from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from typing import Optional, Callable, Any, Type
from uuid import uuid4
from cognite.client.data_classes import Event
from cognite.client.data_classes.data_modeling.instances import InstanceCore
from yaml import safe_dump

from cognite.powerops.clients.powerops_client import get_powerops_client, PowerOpsClient
from cognite.powerops.resync._validation import _clean_relationships
from cognite.powerops.resync.config.resync_config import ReSyncConfig
from cognite.powerops.resync.models.base import Model  # type: ignore[attr-defined]
from cognite.powerops.resync import models
from cognite.powerops.resync._changes import FieldDifference, ModelDifferences
from cognite.powerops.resync.to_models.transform import transform
from cognite.powerops.resync.utils.cdf import get_cognite_api
from cognite.powerops.resync.utils.common import all_concrete_subclasses
from cognite.powerops.utils.cdf import Settings

MODEL_BY_NAME: dict[str, Type[Model]] = {
    model.__name__: model for model in all_concrete_subclasses(Model)  # type: ignore[type-abstract]
}
AVAILABLE_MODELS: frozenset[str] = frozenset(MODEL_BY_NAME)
DEFAULT_MODELS: frozenset[str] = frozenset(
    [m.__name__ for m in [models.ProductionModel, models.MarketModel, models.CogShop1Asset]]
)


def plan(
    path: Path,
    market: str,
    echo: Optional[Callable[[str], None]] = None,
    model_names: Optional[str | list[str]] = None,
    dump_folder: Optional[Path] = None,
    client: PowerOpsClient | None = None,
) -> ModelDifferences:
    echo = echo or print
    settings = Settings()

    client = client or get_powerops_client()

    loaded_models = _load_transform(market, path, client.cdf.config.project, echo, model_names)

    echo(f"Load transform completed, models {', '.join([type(m).__name__ for m in loaded_models])} loaded")
    if settings.powerops.read_dataset is None:
        raise ValueError("No read_dataset configured in settings")
    data_set_external_id = settings.powerops.read_dataset
    all_differences = []
    for new_model in loaded_models:
        echo(f"Retrieving {new_model.model_name} from CDF")
        cdf_model = type(new_model).from_cdf(client, data_set_external_id=data_set_external_id)

        differences = cdf_model.difference(new_model)
        _clean_relationships(client.cdf, differences, new_model, echo)

        if dump_folder:
            dump_folder.mkdir(parents=True, exist_ok=True)
            # Standardize models for easy comparison
            new_model.standardize()
            cdf_model.standardize()

            (dump_folder / f"{new_model.model_name}_local.yaml").write_text(safe_dump(new_model.dump_as_cdf_resource()))
            (dump_folder / f"{new_model.model_name}_cdf.yaml").write_text(safe_dump(cdf_model.dump_as_cdf_resource()))

        all_differences.append(differences)
    return ModelDifferences(all_differences)


def apply(
    path: Path,
    market: str,
    model_names: list[str] | str | None = None,
    echo: Optional[Callable[[Any], None]] = None,
    auto_yes: bool = False,
    echo_pretty: Optional[Callable[[Any], None]] = None,
) -> None:
    echo = echo or print
    echo_pretty = echo_pretty or echo
    client = get_powerops_client()

    loaded_models = _load_transform(market, path, client.cdf.config.project, echo, model_names or list(DEFAULT_MODELS))

    settings = Settings()
    if settings.powerops.read_dataset is None or settings.powerops.write_dataset is None:
        raise ValueError("No read_dataset or write_dataset configured in settings")
    read_dataset = settings.powerops.read_dataset
    retrieved = client.cdf.data_sets.retrieve(external_id=settings.powerops.write_dataset)
    if retrieved is None or retrieved.id is None:
        raise ValueError(f"Could not find write_dataset {settings.powerops.write_dataset}")

    write_dataset = retrieved.id

    for new_model in loaded_models:
        cdf_model = type(new_model).from_cdf(client, data_set_external_id=read_dataset)
        new_sequences_by_id = {s.external_id: s for s in new_model.sequences()}
        new_files_by_id = {f.external_id: f for f in new_model.files()}

        differences = cdf_model.difference(new_model)
        _clean_relationships(client.cdf, differences, new_model, echo)

        for diff in sorted(differences.changes, key=_edges_before_nodes):
            if diff.group == "Domain":
                continue
            if len(diff.unchanged) == diff.total:
                echo(f"No changes detected for {diff.name} in {new_model.model_name}")
                continue
            elif diff.name == "timeseries":
                echo("Found timeseries changes, skipping. These are not updated by resync")
                continue
            echo(f"Changes detected for {diff.name} in {new_model.model_name}")
            echo_pretty(diff.as_summary())
            if diff.changed:
                echo(f"Sample change for {diff.changed[0].changed_fields}")
            ans = "y" if auto_yes else input("Continue? (y/n)")
            if ans.lower() != "y":
                echo("Aborting")
                continue
            diff.set_set_dataset(write_dataset)
            api = get_cognite_api(client.cdf, diff.name, new_sequences_by_id, new_files_by_id)

            if diff.added:
                api.create(diff.added)
                echo(f"Created {len(diff.added)} of {diff.name}")
            if diff.removed:
                api.delete(
                    [r.as_id() if isinstance(r, InstanceCore) else r.external_id for r in diff.removed if r.external_id]
                )
                echo(f"Deleted {len(diff.removed)} of {diff.name}")
            if diff.changed:
                updates = [c.new for c in diff.changed if not c.is_changed_content]
                if updates:
                    api.upsert(updates, mode="replace")
                    echo(f"Updated {len(updates)} of {diff.name}")
                content_updates = [c.new for c in diff.changed if c.is_changed_content]
                if content_updates:
                    api.delete([c.external_id for c in content_updates if c.external_id])
                    api.create(content_updates)
                    echo(f"Updated {len(content_updates)} of {diff.name} with content")


def _edges_before_nodes(diff: FieldDifference) -> int:
    if diff.group == "Domain":
        return 0
    elif diff.group == "CDF" and diff.name not in {"edges", "nodes"}:
        return 1
    elif diff.group == "CDF" and diff.name == "edges":
        return 2
    elif diff.group == "Local" and diff.name == "nodes":
        return 3
    else:
        return 4


def _load_transform(
    market: str, path: Path, cdf_project: str, echo: Callable[[str], None], model_names: str | list[str] | None
) -> list[Model]:
    if isinstance(model_names, str):
        model_names = [model_names]
    elif model_names is None:
        model_names = list(DEFAULT_MODELS)
    elif isinstance(model_names, list) and model_names and isinstance(model_names[0], str):
        model_names = model_names
    else:
        raise ValueError(f"Invalid model_names type: {type(model_names)}")

    if invalid := set(model_names) - AVAILABLE_MODELS:
        raise ValueError(f"Invalid model names: {invalid}. Available models: {AVAILABLE_MODELS}")

    echo(f"Loading and transforming {', '.join(model_names)}")

    config = ReSyncConfig.from_yamls(path, cdf_project)

    return transform(config, market, {MODEL_BY_NAME[model_name] for model_name in model_names})


def _create_bootstrap_finished_event(echo: Callable[[str], None]) -> Event:
    """Creating a POWEROPS_BOOTSTRAP_FINISHED Event in CDF to signal that bootstrap scripts have been ran"""
    current_time = int(datetime.now(timezone.utc).timestamp() * 1000)  # in milliseconds
    event = Event(
        start_time=current_time,
        end_time=current_time,
        external_id=f"POWEROPS_BOOTSTRAP_FINISHED_{str(uuid4())}",
        type="POWEROPS_BOOTSTRAP_FINISHED",
        source="PowerOps bootstrap",
        description="Manual run of bootstrap scripts finished",
    )
    echo(f"Created status event '{event.external_id}'")

    return event


if __name__ == "__main__":
    demo_data = Path(__file__).parent.parent.parent.parent / "tests" / "test_unit" / "test_bootstrap" / "data" / "demo"

    apply(demo_data, "DayAhead", echo=print)



from __future__ import annotations

import json

import pandas as pd
from cognite.client.data_classes import Sequence

from cognite.powerops.clients.data_classes import (
    BenchmarkBidApply,
    BenchmarkProcesApply,
    BidMatrixGeneratorApply,
    DayAheadBidApply,
    DayAheadProcesApply,
    NordPoolMarketApply,
    PriceAreaApply,
    ProductionPlanTimeSeriesApply,
    ReserveScenarioApply,
    RKOMBidApply,
    RKOMBidCombinationApply,
    RKOMCombinationBidApply,
    RKOMMarketApply,
    RKOMProcesApply,
    ScenarioMappingApply,
)
from cognite.powerops.resync.config.market import BenchmarkingConfig, PriceScenario, PriceScenarioID
from cognite.powerops.resync.config.market.dayahead import BidMatrixGeneratorConfig, BidProcessConfig
from cognite.powerops.resync.config.market.market import MARKET_BY_PRICE_AREA
from cognite.powerops.resync.config.market.rkom import RKOMBidCombinationConfig, RKOMBidProcessConfig, RkomMarketConfig
from cognite.powerops.resync.config.resync_config import MarketConfig
from cognite.powerops.resync.models import MarketModel
from cognite.powerops.resync.models import market as market_models
from cognite.powerops.resync.models.cdf_resources import CDFSequence
from cognite.powerops.resync.models.market_dm import (
    BenchmarkMarketDataModel,
    DayAheadMarketDataModel,
    RKOMMarketDataModel,
)
from cognite.powerops.resync.models.production import PriceArea
from cognite.powerops.resync.utils.common import make_ext_id

from ._to_instances import (
    _to_date_transformations,
    _to_scenario_mapping,
    _to_shop_transformation,
    _to_input_timeseries_mapping,
)


def to_market_asset_model(config: MarketConfig, price_areas: list[PriceArea], market_name: str) -> MarketModel:
    nord_pool = market_models.NordPoolMarket(**config.market.model_dump())
    # The external_id is not following the naming convention, so we need to set it manually
    nord_pool.external_id = config.market.external_id

    benchmarking_processes = _to_benchmarking_process(config.benchmarks)

    dayahead_processes = _to_dayahead_process(
        bid_process_configs=config.bidprocess,
        bidmatrix_generators=config.bidmatrix_generators,
        price_scenarios_by_id=config.price_scenario_by_id,
        benchmarking=benchmarking_processes[0],
        price_areas=price_areas,
    )

    rkom = _to_rkom_market(
        rkom_bid_process=config.rkom_bid_process,
        price_scenarios_by_id=config.price_scenario_by_id,
        market_name=market_name,
        rkom_market_config=config.rkom_market,
        rkom_bid_combination_configs=config.rkom_bid_combination,
    )

    model = MarketModel(
        markets=rkom.markets + [nord_pool],
        benchmark_processes=benchmarking_processes,
        dayahead_processes=dayahead_processes,
        rkom_processes=rkom.rkom_processes,
        combinations=rkom.combinations,
    )

    return model


def to_benchmark_data_model(configs: list[BenchmarkingConfig]) -> BenchmarkMarketDataModel:
    if len(configs) > 1:
        # The external id for the benchmarking is hardcoded and thus there can be only one.
        raise NotImplementedError("Only one benchmarking config is supported")
    model = BenchmarkMarketDataModel()
    for config in configs:
        bid = BenchmarkBidApply(
            external_id=make_ext_id(
                [config.bid_date.model_dump_json(), config.market_config_external_id], BenchmarkBidApply
            ),
            date=_to_date_transformations(config.bid_date),
            market=config.market_config_external_id,
        )
        model.bids[bid.external_id] = bid
        process = BenchmarkProcesApply(
            external_id="benchmarking_config_day_ahead",
            name="Benchmarking config DayAhead",
            bid=bid.external_id,
            shop=_to_shop_transformation(config.shop_start, config.shop_end),
            production_plan_time_series=[
                ProductionPlanTimeSeriesApply(
                    external_id=make_ext_id([name, series], ProductionPlanTimeSeriesApply),
                    name=name,
                    series=series,
                )
                for name, series in config.production_plan_time_series
            ],
            run_events=[],
            metrics=config.relevant_shop_objective_metrics,
        )
        model.benchmarking.append(process)
    return model


def to_dayahead_data_model(
    config: MarketConfig,
    dayahead_benchmark: BenchmarkProcesApply,
    benchmark_bid: BenchmarkBidApply,
    price_areas: list[PriceAreaApply],
) -> DayAheadMarketDataModel:
    model = DayAheadMarketDataModel()

    bidmatrix_generators_by_name: dict[str, BidMatrixGeneratorConfig] = {g.name: g for g in config.bidmatrix_generators}

    model.nordpool_market = NordPoolMarketApply(**config.market.model_dump())

    for process in config.bidprocess:
        price_scenarios_by_name = _map_price_scenarios_by_name(
            process.price_scenarios, config.price_scenario_by_id, MARKET_BY_PRICE_AREA[process.price_area_name]
        )

        price_area = next((pa for pa in price_areas if pa.name == process.price_area_name), None)
        if not price_area:
            raise ValueError(f"Price area {process.price_area_name} not found")

        # Create bid matrix generators
        bid_gen_config = bidmatrix_generators_by_name[process.bid_matrix_generator]
        bid_matrix_generators = _to_bid_matrix_generator(bid_gen_config, price_area, process.name)
        model.bid_matrix_generator.update(bid_matrix_generators)

        # Create incremental mapping sequences
        incremental_mappings = []
        for watercourse in price_area.watercourses:
            price_scenarios: dict[str, PriceScenario] = price_scenarios_by_name
            if process.price_scenarios_per_watercourse:
                try:
                    price_scenarios = {
                        scenario_name: price_scenarios[scenario_name]
                        for scenario_name in process.price_scenarios_per_watercourse[watercourse.name]
                    }
                except KeyError as e:
                    raise KeyError(
                        f"Watercourse {watercourse.name} not defined in price_scenarios_per_watercourse "
                        f"for BidProcessConfig {process.name}"
                    ) from e
            for scenario_name, price_scenario in price_scenarios.items():
                external_id = f"SHOP_{watercourse.name}_incremental_mapping_{process.name}_{scenario_name}"
                scenario = _to_scenario_mapping(external_id, scenario_name, price_scenario.to_time_series_mapping())
                incremental_mappings.append(scenario)

        shop = _to_shop_transformation(
            process.shop_start or dayahead_benchmark.shop.start, process.shop_end or dayahead_benchmark.shop.end
        )

        bid = DayAheadBidApply(
            external_id=f"POWEROPS_bid_process_configuration_{process.name}_bid",
            date=_to_date_transformations(process.bid_date or benchmark_bid.date),
            market=model.nordpool_market.external_id,
            is_default_config_for_price_area=process.is_default_config_for_price_area,
            main_scenario=process.main_scenario,
            price_area=f"price_area_{process.price_area_name}",
            # watercourse=process.watercourse,
            price_scenarios=[
                ScenarioMappingApply(
                    external_id=f"SHOP_incremental_mapping_{process.name}_{scenario_name}",
                    mapping_override=[
                        _to_input_timeseries_mapping(entry) for entry in price_scenario.to_time_series_mapping()
                    ],
                    name=scenario_name,
                    shop_type="incremental_mapping",
                )
                for scenario_name, price_scenario in price_scenarios_by_name.items()
            ],
            no_shop=process.no_shop,
            bid_process_configuration_name=process.name,
            bid_matrix_generator_config_external_id=f"POWEROPS_bid_matrix_generator_config_{process.name}",
        )
        model.bids[bid.external_id] = bid

        # Create the DayAheadBidProcess Data Class
        dayahead_process = DayAheadProcesApply(
            name=f"Bid process configuration {process.name}",
            external_id=f"POWEROPS_bid_process_configuration_{process.name}",
            bid=bid.external_id,
            shop=shop,
            bid_matrix_generator_config=list(bid_matrix_generators.values()),
            incremental_mappings=incremental_mappings,
        )

        model.dayahead_processes.append(dayahead_process)

    return model


def _to_bid_matrix_generator(
    bid_gen_config: BidMatrixGeneratorConfig, price_area: PriceAreaApply, process_name: str
) -> dict[str, BidMatrixGeneratorApply]:
    return {
        (
            external_id := f"POWEROPS_bid_matrix_generator_config_{process_name}"
            # make_ext_id(
            #     [plant.name, bid_gen_config.default_method, bid_gen_config.default_function_external_id],
            #     BidMatrixGeneratorApply,
            # )
        ): BidMatrixGeneratorApply(
            external_id=external_id,
            shop_plant=plant.name,
            methods=bid_gen_config.default_method,
            function_external_id=bid_gen_config.default_function_external_id,
        )
        for plant in price_area.plants
    }


def to_rkom_data_model(config: MarketConfig, market_name: str) -> RKOMMarketDataModel:
    model = RKOMMarketDataModel()

    model.rkom_market = RKOMMarketApply(**config.rkom_market.model_dump())

    for process in config.rkom_bid_process:
        price_scenarios_by_name = _map_price_scenarios_by_name(
            process.price_scenarios,
            config.price_scenario_by_id,
            market_name,
        )

        incremental_mappings = []
        reserve_scenarios = []
        scenario_mappings = []
        for scenario_name, price_scenario in price_scenarios_by_name.items():
            external_id = f"SHOP_RKOM_{process.watercourse}_incremental_mapping_{process.name}_{scenario_name}"
            scenario_mapping = _to_scenario_mapping(external_id, scenario_name, price_scenario.to_time_series_mapping())
            scenario_mappings.append(scenario_mapping)
            for reserve_scenario in process.reserve_scenarios.list_scenarios():
                name = (
                    f"{reserve_scenario.auction.name}_{reserve_scenario.block}_{reserve_scenario.product}_"
                    f"{reserve_scenario.reserve_group}_{reserve_scenario.volume}"
                )
                external_id = (
                    f"SHOP_{process.watercourse}_incremental_mapping_"
                    f"{process.name}_{scenario_name}_{reserve_scenario.volume}MW"
                )
                reserve_mapping = _to_scenario_mapping(
                    external_id,
                    name,
                    price_scenario.to_time_series_mapping() + reserve_scenario.to_time_series_mapping(),
                )
                incremental_mappings.append(reserve_mapping)

                external_id = make_ext_id(
                    reserve_mapping.model_dump(exclude={"obligation_external_id", "mip_plant_time_series"}),
                    ReserveScenarioApply,
                )
                apply = ReserveScenarioApply(
                    external_id=external_id,
                    auction=reserve_scenario.auction.name,
                    block=reserve_scenario.block,
                    product=reserve_scenario.product,
                    reserve_group=reserve_scenario.reserve_group,
                    volume=reserve_scenario.volume,
                )
                reserve_scenarios.append(apply)

        bid = RKOMBidApply(
            external_id=f"{process.external_id}_bid",
            date=_to_date_transformations(process.bid_date),
            watercourse=process.watercourse,
            method=process.method,
            minimum_price=process.minimum_price,
            price_premium=process.price_premium,
            price_scenarios=scenario_mappings,
            reserve_scenarios=reserve_scenarios,
            market=model.rkom_market.external_id,
        )
        model.bids[bid.external_id] = bid
        shop = _to_shop_transformation(process.shop_start, process.shop_end)

        apply = RKOMProcesApply(
            external_id=process.external_id,
            name=process.name,
            bid=bid.external_id,
            shop=shop,
            timezone=process.timezone,
            plants=process.rkom_plants,
            incremental_mappings=incremental_mappings,
        )
        model.rkom_processes.append(apply)

    for comb in config.rkom_bid_combination or []:
        sequence_external_id = f"RKOM_bid_combination_configuration_{comb.auction.value}_{comb.name}"

        bid = RKOMCombinationBidApply(
            external_id=make_ext_id(
                [comb.auction.value, comb.name] + comb.rkom_bid_config_external_ids, RKOMCombinationBidApply
            ),
            auction=comb.auction.value,
            name=comb.name,
            rkom_bid_configs=comb.rkom_bid_config_external_ids,
        )

        combination = RKOMBidCombinationApply(
            external_id=sequence_external_id,
            name=sequence_external_id.replace("_", " "),
            bid=bid,
            auction=comb.auction.value,
        )
        model.rkom_bid_combinations.append(combination)

    return model


def _to_benchmarking_process(benchmarks: list[BenchmarkingConfig]) -> list[market_models.BenchmarkProcess]:
    benchmarkings: list[market_models.BenchmarkProcess] = []
    if len(benchmarks) > 1:
        # The external id for the benchmarking is hardcoded and thus there can be only one.
        raise NotImplementedError("Only one benchmarking config is supported")
    for benchmarking_config in benchmarks:
        bid = market_models.BenchmarkBid(
            date=json.dumps(benchmarking_config.bid_date.operations),
            market_config_external_id=benchmarking_config.market_config_external_id,
        )
        process = market_models.BenchmarkProcess(
            name="Benchmarking config DA",
            bid=bid,
            shop=market_models.ShopTransformation(
                starttime=json.dumps(benchmarking_config.shop_start.operations),
                endtime=json.dumps(benchmarking_config.shop_end.operations),
            ),
            production_plan_time_series=json.dumps(benchmarking_config.production_plan_time_series, ensure_ascii=False)
            if benchmarking_config.production_plan_time_series
            else [],  # ensure_ascii=False to treat Nordic letters properly,
            benchmarking_metrics=benchmarking_config.relevant_shop_objective_metrics,
        )
        # The external_id is not following the naming convention, so we need to set it manually
        process.external_id = "POWEROPS_dayahead_bidding_benchmarking_config"
        benchmarkings.append(process)
    return benchmarkings


def _to_dayahead_process(
    bid_process_configs: list[BidProcessConfig],
    bidmatrix_generators: list[BidMatrixGeneratorConfig],
    price_scenarios_by_id: dict[str, PriceScenario],
    benchmarking: market_models.BenchmarkProcess,
    price_areas: list[PriceArea],
) -> list[market_models.DayAheadProcess]:
    dayahead_processes: list[market_models.DayAheadProcess] = []
    bidmatrix_generators_by_name: dict[str, BidMatrixGeneratorConfig] = {g.name: g for g in bidmatrix_generators}
    for process in bid_process_configs:
        price_scenarios_by_name = _map_price_scenarios_by_name(
            process.price_scenarios, price_scenarios_by_id, MARKET_BY_PRICE_AREA[process.price_area_name]
        )

        price_area = next((pa for pa in price_areas if pa.name == process.price_area_name), None)
        if not price_area:
            raise ValueError(f"Price area {process.price_area_name} not found")

        # Create bid matrix generator config sequence
        bidmatrix_generator = bidmatrix_generators_by_name[process.bid_matrix_generator]
        column_def = [
            {"valueType": "STRING", "externalId": external_id}
            for external_id in bidmatrix_generator.column_external_ids
        ]
        bid_matrix_generator_sequence = Sequence(
            external_id=f"POWEROPS_bid_matrix_generator_config_{process.name}",
            name=f"POWEROPS bid matrix generator config {process.name}",
            description="Configuration of bid matrix generation method to use for each plant in the price area",
            columns=column_def,
            metadata={
                # TODO: Rename this from "shop:type" to something without "shop"
                #   (but check if e.g. power-ops-functions uses this)
                "bid:price_area": f"price_area_{process.price_area_name}",
                "shop:type": "bid_matrix_generator_config",
                "bid:bid_process_configuration_name": process.name,
            },
        )
        bid_matrix_generator_sequence_content = pd.DataFrame(
            [
                [plant.name, bidmatrix_generator.default_method, bidmatrix_generator.default_function_external_id]
                for plant in price_area.plants
            ],
            columns=bidmatrix_generator.column_external_ids,
        )

        # Create incremental mapping sequences
        incremental_mapping_sequences = []
        for watercourse in price_area.watercourses:
            price_scenarios: dict[str, PriceScenario] = price_scenarios_by_name
            if process.price_scenarios_per_watercourse:
                try:
                    price_scenarios = {
                        scenario_name: price_scenarios[scenario_name]
                        for scenario_name in process.price_scenarios_per_watercourse[watercourse.name]
                    }
                except KeyError as e:
                    raise KeyError(
                        f"Watercourse {watercourse.name} not defined in price_scenarios_per_watercourse "
                        f"for BidProcessConfig {process.name}"
                    ) from e
            for scenario_name, price_scenario in price_scenarios.items():
                time_series_mapping = price_scenario.to_time_series_mapping()
                metadata = {
                    "shop:watercourse": watercourse.name,
                    "shop:type": "incremental_mapping",
                    "bid:scenario_name": scenario_name,
                }
                external_id = f"SHOP_{watercourse.name}_incremental_mapping_{process.name}_{scenario_name}"
                name = external_id.replace("_", " ")
                sequence = Sequence(
                    name=name,
                    external_id=external_id,
                    description="Mapping between SHOP paths and CDF TimeSeries",
                    columns=time_series_mapping.column_definitions,
                    metadata=metadata,
                )
                incremental_mapping = CDFSequence(sequence=sequence, content=time_series_mapping.to_dataframe())
                incremental_mapping_sequences.append(incremental_mapping)

        # Create the DayAheadBidProcess Data Class
        dayahead_process = market_models.DayAheadProcess(
            name=f"Bid process configuration {process.name}",
            description="Bid process configuration defining how to run a bid matrix generation process",
            bid=market_models.DayAheadBid(
                date=str(process.bid_date or benchmarking.bid.date),
                market_config_external_id=benchmarking.bid.market_config_external_id,
                is_default_config_for_price_area=process.is_default_config_for_price_area,
                main_scenario=process.main_scenario,
                price_area=f"price_area_{process.price_area_name}",
                price_scenarios={
                    scenario_name: scenario.time_series_external_id
                    for scenario_name, scenario in price_scenarios_by_name.items()
                },
                no_shop=process.no_shop,
                bid_process_configuration_name=process.name,
                bid_matrix_generator_config_external_id=f"POWEROPS_bid_matrix_generator_config_{process.name}",
            ),
            shop=market_models.ShopTransformation(
                starttime=str(process.shop_start or benchmarking.shop.starttime),
                endtime=str(process.shop_end or benchmarking.shop.endtime),
            ),
            bid_matrix_generator_config=CDFSequence(
                sequence=bid_matrix_generator_sequence,
                content=bid_matrix_generator_sequence_content,
            ),
            incremental_mapping=incremental_mapping_sequences,
        )
        # The IDs are inconsistently compared to the other data classes, so we need to set them manually
        dayahead_process.external_id = f"POWEROPS_bid_process_configuration_{process.name}"
        dayahead_process.type_ = "POWEROPS_bid_process_configuration"

        dayahead_processes.append(dayahead_process)

    return dayahead_processes


def _to_rkom_market(
    rkom_bid_process: list[RKOMBidProcessConfig],
    price_scenarios_by_id: dict[str, PriceScenario],
    market_name: str,
    rkom_market_config: RkomMarketConfig,
    rkom_bid_combination_configs: list[RKOMBidCombinationConfig] | None = None,
) -> MarketModel:
    model = MarketModel()

    rkom_market = market_models.RKOMMarket(**rkom_market_config.model_dump())
    rkom_market.external_id = rkom_market_config.external_id
    model.markets.append(rkom_market)

    for config in rkom_bid_process:
        price_scenarios_by_name = _map_price_scenarios_by_name(
            config.price_scenarios,
            price_scenarios_by_id,
            market_name,
        )
        incremental_mapping_sequences = []
        for scenario_name, price_scenario in price_scenarios_by_name.items():
            mappings = price_scenario.to_time_series_mapping()
            for reserve_scenario in config.reserve_scenarios:
                reserve_mapping = reserve_scenario.to_time_series_mapping()
                metadata = {
                    "shop:watercourse": config.watercourse,
                    "shop:type": "rkom_incremental_mapping",
                    "bid:scenario_name": scenario_name,
                    "bid:reserve_volume": str(reserve_scenario.volume),
                }
                external_id = (
                    f"SHOP_{config.watercourse}_incremental_mapping_"
                    f"{config.name}_{scenario_name}_{reserve_scenario.volume}MW"
                )
                name = f"{config.watercourse} {scenario_name} {reserve_scenario.volume} MW"
                sequence = Sequence(
                    name=name,
                    external_id=external_id,
                    description="Mapping between SHOP paths and CDF TimeSeries",
                    columns=(mappings + reserve_mapping).column_definitions,
                    metadata=metadata,
                )
                content = (mappings + reserve_mapping).to_dataframe()
                incremental_mapping = CDFSequence(sequence=sequence, content=content)
                incremental_mapping_sequences.append(incremental_mapping)

        process = market_models.RKOMProcess(
            name=config.name,
            description=f"RKOM bid generation config for {config.watercourse}",
            bid=market_models.RKOMBid(
                date=str(config.bid_date),
                auction=config.reserve_scenarios.auction.value,
                block=config.reserve_scenarios.block,
                product=config.reserve_scenarios.product,
                watercourse=config.watercourse,
                method=config.method,
                minimum_price=str(config.minimum_price),
                price_premium=str(config.price_premium),
                price_scenarios=json.dumps(list(price_scenarios_by_name)),
                reserve_scenarios=str(config.reserve_scenarios),
            ),
            shop=market_models.ShopTransformation(
                starttime=str(config.shop_start),
                endtime=str(config.shop_end),
            ),
            timezone=config.timezone,
            rkom=market_models.RKOMPlants(plants=json.dumps(sorted(config.rkom_plants))),
            incremental_mapping=incremental_mapping_sequences,
        )
        process.external_id = config.external_id
        process.type_ = "POWEROPS"
        model.rkom_processes.append(process)

    for config in rkom_bid_combination_configs or []:
        sequence_external_id = f"RKOM_bid_combination_configuration_{config.auction.value}_{config.name}"
        combination = market_models.RKOMBidCombination(
            name=sequence_external_id.replace("_", " "),
            description="Defining which RKOM bid methods should be combined (into the total bid form)",
            bid=market_models.RKOMCombinationBid(
                auction=config.auction.value,
                combination_name=config.name,
                rkom_bid_configs=config.rkom_bid_config_external_ids,
            ),
        )
        combination.external_id = sequence_external_id
        model.combinations.append(combination)

    return model


def _map_price_scenarios_by_name(
    scenario_ids: list[PriceScenarioID], price_scenarios_by_id: dict[str, PriceScenario], market_name: str
) -> dict[str, PriceScenario]:
    scenario_by_name = {}
    for identifier in scenario_ids:
        ref_scenario = price_scenarios_by_id[identifier.id]
        name = identifier.rename or ref_scenario.name or identifier.id
        scenario_by_name[name] = PriceScenario(name=market_name, **ref_scenario.model_dump(exclude={"name"}))
    return scenario_by_name



from .asset_model import AssetModel
from .asset_type import AssetType, NonAssetType
from .data_model import DataModel
from .model import Model, T_Model

__all__ = ["AssetModel", "AssetType", "NonAssetType", "DataModel", "Model", "T_Model"]



from __future__ import annotations

import logging
import random

from cognite.powerops.clients._api_client import CogShopAPIs

logger = logging.getLogger(__name__)


class ShopModel:
    def __init__(self) -> None:
        self.model_id = random.randint(1000, 9999)

    def render_yaml(self) -> str:
        return "sintef_shop_model_yaml_representation"

    def update(self):
        raise NotImplementedError


class ShopModelsAPI:
    def __init__(self, cogshop: CogShopAPIs):
        self._cogshop = cogshop

    def list(self) -> list[ShopModel]:
        return [ShopModel()]

    def retrieve(self, model_id):
        m = ShopModel()
        m.model_id = model_id
        return m



from __future__ import annotations

from pydantic import Field
from typing import Type
from cognite.powerops.clients.data_classes import (
    GeneratorApply,
    PlantApply,
    PriceAreaApply,
    ReservoirApply,
    WatercourseApply,
)
from cognite.powerops.clients.powerops_client import PowerOpsClient
from cognite.powerops.resync.models.base import DataModel, T_Model
from cognite.powerops.resync.models.cdf_resources import CDFSequence


class ProductionModelDM(DataModel):
    cdf_sequences: list[CDFSequence] = Field(default_factory=list)
    price_areas: list[PriceAreaApply] = Field(default_factory=list)
    watercourses: list[WatercourseApply] = Field(default_factory=list)
    plants: list[PlantApply] = Field(default_factory=list)
    generators: list[GeneratorApply] = Field(default_factory=list)
    reservoirs: list[ReservoirApply] = Field(default_factory=list)

    @classmethod
    def from_cdf(
        cls: Type[T_Model], client: PowerOpsClient, fetch_metadata: bool = True, fetch_content: bool = False
    ) -> T_Model:
        ...



from __future__ import annotations

from itertools import product
from typing import Type, ClassVar

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import ContainerId
from pydantic import Field, field_validator

from cognite.powerops.clients.data_classes import (
    ScenarioTemplateApply,
    ValueTransformationApply,
    OutputContainerApply,
    ScenarioMappingApply,
)
import cognite.powerops.cogshop1.data_classes as cogshop_v1
from .base import DataModel, Model
from .cdf_resources import CDFFile, CDFSequence
from cognite.powerops.cogshop1.data_classes._core import DomainModelApply as DomainModelApplyCogShop1
from ...clients import PowerOpsClient

ExternalID = str


class CogShopCore(Model):
    shop_files: list[CDFFile] = Field(default_factory=list)

    @field_validator("shop_files", mode="after")
    def ordering(cls, value: list[CDFFile]) -> list[CDFFile]:
        return sorted(value, key=lambda x: x.external_id)

    def standardize(self) -> None:
        self.shop_files = self.ordering(self.shop_files)


class CogShopDataModel(CogShopCore, DataModel):
    scenario_templates: list[ScenarioTemplateApply] = Field(default_factory=list)
    base_mappings: dict[ExternalID, ScenarioMappingApply] = Field(default_factory=dict)
    output_definitions: dict[ExternalID, OutputContainerApply] = Field(default_factory=dict)
    value_transformations: dict[ExternalID, ValueTransformationApply] = Field(default_factory=dict)


class CogShop1Asset(CogShopCore, DataModel, protected_namespaces=()):
    cls_by_container: ClassVar[dict[ContainerId, Type[DomainModelApplyCogShop1]]] = {
        ContainerId("cogShop", "ModelTemplate"): cogshop_v1.ModelTemplateApply,
        ContainerId("cogShop", "Mapping"): cogshop_v1.MappingApply,
        ContainerId("cogShop", "FileRef"): cogshop_v1.FileRefApply,
        ContainerId("cogShop", "Transformation"): cogshop_v1.TransformationApply,
    }
    model_templates: dict[ExternalID, cogshop_v1.ModelTemplateApply] = Field(default_factory=dict)
    mappings: dict[ExternalID, cogshop_v1.MappingApply] = Field(default_factory=dict)
    transformations: dict[ExternalID, cogshop_v1.TransformationApply] = Field(default_factory=dict)
    base_mappings: list[CDFSequence] = Field(default_factory=list)
    output_definitions: list[CDFSequence] = Field(default_factory=list)

    @field_validator("base_mappings", mode="after")
    def ordering_sequences(cls, value: list[CDFSequence]) -> list[CDFSequence]:
        return sorted(value, key=lambda x: x.external_id)

    @field_validator("model_templates", "mappings", "transformations", mode="after")
    def ordering_dict(cls, value: dict) -> dict:
        return {k: v for k, v in sorted(value.items(), key=lambda x: x[0])}

    @classmethod
    def from_cdf(cls, client: PowerOpsClient, data_set_external_id: str) -> CogShop1Asset:
        cog_shop = client.cog_shop1
        templates = cog_shop.model_templates.list(limit=-1)
        base_mappings = cog_shop.mappings.list(limit=-1)
        transformations = cog_shop.transformations.list(limit=-1)
        files = cog_shop.file_refs.list(limit=-1)

        transformation_by_id: dict[str, cogshop_v1.TransformationApply] = {t.external_id: t for t in transformations}
        mappings_by_id = {}
        readme_fields = {"created_time", "deleted_time", "last_updated_time", "version"}
        for transformation in transformations:
            data = transformation.model_dump(exclude=readme_fields)
            apply = cogshop_v1.TransformationApply(**data)
            transformation_by_id[apply.external_id] = apply

        for mapping in base_mappings:
            data = mapping.model_dump(exclude=readme_fields)
            data["transformations"] = sorted(
                [transformation_by_id[t] for t in data["transformations"]], key=lambda x: x.order
            )
            apply = cogshop_v1.MappingApply(**data)
            mappings_by_id[apply.external_id] = apply

        file_by_id = {f.external_id: f for f in files}
        model_templates = {}
        for template in templates:
            data = template.model_dump(exclude=readme_fields - {"version"})
            data["version"] = str(data["version"])
            if data["model"] in file_by_id:
                data["model"] = file_by_id[data["model"]].model_dump(exclude=readme_fields)
            data["base_mappings"] = [mappings_by_id[m] for m in data["base_mappings"]]
            apply = cogshop_v1.ModelTemplateApply(**data)
            model_templates[apply.external_id] = apply

        # There files and sequences are not linked to the model templates
        # (this should be done in the next version of Cogshop).
        watercourse_names = list({t.watercourse for t in model_templates.values()})
        sequence_ids = [
            f"SHOP_{name}{suffix}"
            for name, suffix in product(watercourse_names, ["_base_mapping", "_output_definition"])
        ]
        cdf_client: CogniteClient = client.cdf
        if sequence_ids:
            sequences = cdf_client.sequences.retrieve_multiple(external_ids=sequence_ids)
        else:
            all_sequences = cdf_client.sequences.list(limit=-1, external_id_prefix="SHOP_")
            sequences = [
                s
                for s in all_sequences
                if s.external_id.endswith("_base_mapping") or s.external_id.endswith("_output_definition")
            ]

        base_mappings = [CDFSequence(sequence=s) for s in sequences if s.external_id.endswith("_base_mapping")]
        output_definitions = [
            CDFSequence(sequence=s) for s in sequences if s.external_id.endswith("_output_definition")
        ]

        files = cdf_client.files.list(
            limit=-1,
            source="PowerOps bootstrap",
            mime_type="text/plain",
        )
        shop_files = [CDFFile(meta=f) for f in files]

        return cls(
            model_templates=model_templates,
            mappings=mappings_by_id,
            transformations=transformation_by_id,
            base_mappings=base_mappings,
            output_definitions=output_definitions,
            shop_files=shop_files,
        )

    def standardize(self) -> None:
        super().standardize()
        self.base_mappings = self.ordering_sequences(self.base_mappings)
        self.output_definitions = self.ordering_sequences(self.output_definitions)
        self.model_templates = self.ordering_dict(self.model_templates)
        self.mappings = self.ordering_dict(self.mappings)
        self.transformations = self.ordering_dict(self.transformations)
        for template in self.model_templates.values():
            template.base_mappings = sorted(template.base_mappings, key=lambda x: x.external_id)

        for mapping in self.mappings.values():
            mapping.transformations = sorted(mapping.transformations, key=lambda x: x.order)



from __future__ import annotations
import json

from dataclasses import dataclass
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Optional, ClassVar, Callable
import traceback


from cognite.client._constants import MAX_VALID_INTERNAL_ID
from cognite.client import CogniteClient
from cognite.client.data_classes import ExtractionPipeline, ExtractionPipelineRun
from cognite.client.exceptions import CogniteAPIError

MSG_CHAR_LIMIT = 1000


class RunStatus(str, Enum):
    SUCCESS = "success"
    FAILURE = "failure"
    SEEN = "seen"


@dataclass
class _Config:
    dump_truncated_to_file: bool = True
    truncate_keys: list[str] = None
    log_file_prefix: str | None = None


class PipelineRun:
    """
    A context manager to create an extraction pipeline run when exiting the context.
    Run status defaults to SEEN and FAILURE if an exception is raised. To set a custom
    status and message, use the `update_data` method.

    """

    init_status: ClassVar[RunStatus] = RunStatus.SEEN
    log_file_external_id: ClassVar[str] = "log_file_external_id"
    log_file_id: ClassVar[str] = "log_file_id"
    exception: ClassVar[str] = "exception"

    reserved_keys: ClassVar[tuple[str]] = [log_file_external_id, log_file_id, exception]

    def __init__(
        self,
        client: CogniteClient,
        pipeline_external_id: str,
        config: _Config,
        data_set_id: int,
        error_logger: Callable[[str], None],
    ) -> None:
        self.client = client
        self.pipeline_external_id = pipeline_external_id
        self.config = config
        self.data_set_id = data_set_id
        self.error_logger = error_logger

        self.data = {}
        self.status = self.init_status

    def __enter__(self) -> "PipelineRun":
        return self

    def update_data(
        self,
        status: RunStatus = init_status,
        **data: Any,
    ) -> "PipelineRun":
        self.status = status
        if isinstance(data, dict):
            for key in self.reserved_keys:
                if key in data:
                    self.error_logger(f"Key {key} is reserved and cannot be used in data. This will be skipped")
                    del data[key]
            self.data.update(data)
        else:
            raise TypeError(f"Data must be a dict, got {type(data)}")
        return self

    def __exit__(self, exc_type=None, exc_value=None, traceback_=None) -> bool:
        suppress_exception = False

        if any((exc_type, exc_value, traceback_)):
            self.status = RunStatus.FAILURE
            self.data[self.exception] = traceback.format_exc()

        message = self.get_message(self.config.dump_truncated_to_file)
        run = ExtractionPipelineRun(
            status=self.status.value,
            extpipe_external_id=self.pipeline_external_id,
            message=message,
        )
        self.client.extraction_pipelines.runs.create(run)

        return suppress_exception

    def get_message(self, dump_truncated_to_file: bool = None) -> str:
        file_external_id = (
            f"{self.config.log_file_prefix}/{self.pipeline_external_id}/"
            f"{datetime.now(timezone.utc).isoformat().replace(':', '_')}"
        )
        data, file_content = self._create_run_data_and_file_content(file_external_id, dump_truncated_to_file)
        if dump_truncated_to_file and file_content:
            try:
                file_id = self.client.files.upload_bytes(
                    content=file_content,
                    name=f"{file_external_id.lower()}.log".replace("/", "_"),
                    external_id=file_external_id,
                    mime_type="text/plain",
                    data_set_id=self.data_set_id,
                ).id
            except CogniteAPIError as e:
                self.error_logger(f"Failed to upload log file {file_external_id}: {e}")
                file_id = ""

            data[self.log_file_id] = file_id

        return self._as_json(data)

    def _create_run_data_and_file_content(
        self, file_external_id: str, dump_truncated_to_file: bool
    ) -> tuple[dict, str]:
        file_content = []
        data = self.data.copy()
        if dump_truncated_to_file:
            data[self.log_file_external_id] = file_external_id
            # Need a dummy id as we do not know the id before the file is uploaded
            data[self.log_file_id] = MAX_VALID_INTERNAL_ID

        dumped = self._as_json(data)
        if (above_limit := len(dumped) - MSG_CHAR_LIMIT) > 0:
            # In case, truncating the specified keys is not enough, we also start to truncate everything else.
            truncate_keys = self.config.truncate_keys + list(
                set(data) - {self.log_file_id, self.log_file_external_id} - set(self.config.truncate_keys)
            )
            for key in truncate_keys:
                if key not in data:
                    continue
                file_content.append(
                    f"{'='*70}\n{key}\n{'='*70}\n{self.data[key]}\n{'='*70}",
                )
                entry_length = len(data[key])
                if entry_length >= above_limit + 3:
                    data[key] = data[key][: entry_length - (above_limit + 3)] + "..."
                    break
                elif entry_length < 3:
                    continue
                else:
                    reduction = min((above_limit + 3) - entry_length, entry_length)
                    data[key] = "..."
                    above_limit = above_limit - reduction

        return data, "\n\n".join(file_content)

    @staticmethod
    def _as_json(data: dict) -> str:
        # Removing space to use as little space as possible
        return json.dumps(data, separators=(",", ":"))


class ExtractionPipelineCreate:
    """
    This is a simplified write version of an Extraction pipeline.

    This ensures that the Extraction Pipeline exists when calling the `create_pipeline_run` method.

    """

    def __init__(
        self,
        external_id: str,
        data_set_external_id: str,
        description: str | None = None,
        dump_truncated_to_file: bool = True,
        truncate_keys: list[str] = None,
        log_file_prefix: str | None = None,
    ) -> None:
        self.external_id = external_id
        self.dataset_external_id = data_set_external_id
        self.description = description
        self._data_set_id: Optional[int] = None

        self.config = _Config(
            dump_truncated_to_file=dump_truncated_to_file,
            truncate_keys=truncate_keys,
            log_file_prefix=log_file_prefix,
        )

    def get_or_create(self, client: CogniteClient):
        extraction_pipeline = client.extraction_pipelines.retrieve(external_id=self.external_id)
        if extraction_pipeline is None:
            self._data_set_id = client.data_sets.retrieve(external_id=self.dataset_external_id).id
            client.extraction_pipelines.create(
                ExtractionPipeline(
                    external_id=self.external_id,
                    name=self.external_id,
                    data_set_id=self._data_set_id,
                    description=self.description,
                )
            )
        return self

    def create_pipeline_run(self, client: CogniteClient, error_logger: Callable[[str], None] = None) -> PipelineRun:
        """

        Args:
            client: An instance of the CogniteClient.
            error_logger: The function to use for logging errors. Defaults to print.

        Returns:
            Return a pipeline run that can be used as a context manager.
        """
        self._data_set_id = self._data_set_id or client.data_sets.retrieve(external_id=self.dataset_external_id).id

        return PipelineRun(client, self.external_id, self.config, self._data_set_id, error_logger or print)



from __future__ import annotations

import abc
from abc import ABC
from typing import ClassVar, Callable, Iterable, Type as TypingType, Any, TypeVar, Literal

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteResource, CogniteResourceList
from cognite.client.data_classes import (
    TimeSeriesList,
    SequenceUpdate,
    FileMetadataUpdate,
)
from typing_extensions import Self

from cognite.client.data_classes import TimeSeries
from pydantic import BaseModel
from cognite.powerops.clients.powerops_client import PowerOpsClient
from cognite.powerops.resync._changes import Change, FieldDifference, ModelDifference
from cognite.powerops.resync.models.base.resource_type import _T_Type, ResourceType
from cognite.powerops.resync.models.cdf_resources import CDFSequence, CDFFile, CDFResource
from cognite.powerops.resync.models.helpers import isinstance_list
from cognite.powerops.resync.utils.serializer import remove_read_only_fields
from cognite.powerops.clients.data_classes._core import DomainModelApply
from cognite.powerops.cogshop1.data_classes._core import DomainModelApply as DomainModelApplyCogShop1


class Model(BaseModel, ABC):
    def sequences(self) -> list[CDFSequence]:
        sequences = [sequence for item in self._resource_types() for sequence in item.sequences()]
        sequences.extend(self._fields_of_type(CDFSequence))
        return sequences

    def files(self) -> list[CDFFile]:
        files = [file for item in self._resource_types() for file in item.files()]
        files.extend(self._fields_of_type(CDFFile))
        return files

    def timeseries(self) -> TimeSeriesList:
        time_series = [ts for item in self._resource_types() for ts in item.time_series()]
        time_series.extend(self._fields_of_type(TimeSeries))
        return TimeSeriesList(time_series)

    cdf_resources: ClassVar[dict[Callable, type]] = {
        sequences: CDFSequence,
        files: CDFFile,
        timeseries: TimeSeries,
    }

    def _resource_types(self) -> Iterable[ResourceType]:
        yield from self._fields_of_type(ResourceType)

    def _fields_of_type(self, type_: TypingType[_T_Type]) -> Iterable[_T_Type]:
        for field_name in self.model_fields:
            value = getattr(self, field_name)
            if isinstance(value, type_):
                yield value
            elif isinstance_list(value, type_):
                yield from value

    @property
    def model_name(self) -> str:
        return type(self).__name__

    def dump_as_cdf_resource(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        for resource_fun in self.cdf_resources.keys():
            name = resource_fun.__name__
            if items := resource_fun(self):

                def dump(resource: Any) -> dict[str, Any]:
                    return remove_read_only_fields(resource.dump(camel_case=True))

                output[name] = [dump(item) for item in items]
        return output

    @classmethod
    def _load_by_type_external_id(cls, data: dict[str, Any]) -> dict[str, dict[str, Any]]:
        loaded_by_type_external_id: dict[str, dict[str, Any]] = {}
        for function, resource_cls in cls.cdf_resources.items():
            name = function.__name__
            if items := data.get(name):
                loaded_by_type_external_id[name] = {
                    loaded.external_id: loaded
                    for loaded in (resource_cls._load(item) if isinstance(item, dict) else item for item in items)
                }
        return loaded_by_type_external_id

    @classmethod
    @abc.abstractmethod
    def load_from_cdf_resources(cls: TypingType[Self], data: dict[str, Any]) -> Self:
        raise NotImplementedError()

    @classmethod
    def _external_id_key(cls, resource: dict | Any) -> str:
        if hasattr(resource, "external_id"):
            return resource.external_id.casefold()
        elif isinstance(resource, dict) and ("external_id" in resource or "externalId" in resource):
            return resource.get("external_id", resource.get("externalId")).casefold()
        raise ValueError(f"Could not find external_id in {resource}")

    @classmethod
    @abc.abstractmethod
    def from_cdf(
        cls: TypingType[T_Model],
        client: PowerOpsClient,
        data_set_external_id: str,
    ) -> T_Model:
        ...

    @abc.abstractmethod
    def standardize(self) -> None:
        """
        This ensures that the model is in a standardized form.

        This is useful when doing comparisons between models; as for example, the ordering of the assets in some
        lists does not matter.
        """
        raise NotImplementedError()

    def difference(self, new_model: T_Model) -> ModelDifference:
        if type(self) != type(new_model):
            raise ValueError(f"Cannot compare model of type {type(self)} with {type(new_model)}")
        # The dump and load calls are to remove all read only fields
        current_reloaded = self.load_from_cdf_resources(self.dump_as_cdf_resource())
        new_reloaded = new_model.load_from_cdf_resources(new_model.dump_as_cdf_resource())

        diffs = []
        for field_name in self.model_fields:
            current_value = getattr(current_reloaded, field_name)
            new_value = getattr(new_reloaded, field_name)
            if type(current_value) != type(new_value):
                raise ValueError(
                    f"Cannot compare field {field_name} of type {type(current_value)} with {type(new_value)}"
                )
            diff = self._find_diffs(current_value, new_value, "Domain", field_name)
            diffs.append(diff)

        for function in self.cdf_resources:
            current_items = function(current_reloaded)
            new_items = function(new_reloaded)

            diff = self._find_diffs(
                current_items,
                new_items,
                "CDF",
                function.__name__,
            )

            diffs.append(diff)

        return ModelDifference(name=self.model_name, changes=diffs)

    @staticmethod
    def _find_diffs(
        current_value: Any, new_value: Any, group: Literal["CDF", "Domain"], field_name: str
    ) -> FieldDifference:
        if not current_value and not new_value:
            return FieldDifference(group=group, name=field_name)

        if (
            isinstance(current_value, (list, CogniteResourceList))
            and current_value
            and isinstance(current_value[0], (ResourceType, CogniteResource))
        ) or (
            isinstance(new_value, (list, CogniteResourceList))
            and new_value
            and isinstance(new_value[0], (ResourceType, CogniteResource))
        ):
            current_value_by_id = {item.external_id: item for item in current_value}
            new_value_by_id = {item.external_id: item for item in new_value}
        elif (
            isinstance(current_value, list) and current_value and isinstance(current_value[0], (CDFSequence, CDFFile))
        ) or (isinstance(new_value, list) and new_value and isinstance(new_value[0], (CDFSequence, CDFFile))):
            current_value_by_id = {item.external_id: item.cdf_resource for item in current_value}
            new_value_by_id = {item.external_id: item.cdf_resource for item in new_value}
        elif (
            isinstance(current_value, dict)
            and current_value
            and isinstance(next(iter(current_value.values())), (DomainModelApply, DomainModelApplyCogShop1))
        ) or (
            isinstance(new_value, dict)
            and new_value
            and isinstance(next(iter(new_value.values())), (DomainModelApply, DomainModelApplyCogShop1))
        ):
            current_value_by_id = {item.external_id: item for item in current_value.values()}
            new_value_by_id = {item.external_id: item for item in new_value.values()}
        else:
            raise NotImplementedError(f"{type(current_value)} is not supported")

        added_ids = set(new_value_by_id.keys()) - set(current_value_by_id.keys())
        added = [new_value_by_id[external_id] for external_id in sorted(added_ids)]

        removed_ids = set(current_value_by_id.keys()) - set(new_value_by_id.keys())
        removed = [current_value_by_id[external_id] for external_id in sorted(removed_ids)]

        existing_ids = set(current_value_by_id.keys()) & set(new_value_by_id.keys())
        changed = []
        unchanged = []
        for existing in sorted(existing_ids):
            current = current_value_by_id[existing]
            new = new_value_by_id[existing]
            if current == new:
                unchanged.append(current)
            else:
                changed.append(Change(last=current, new=new))

        return FieldDifference(
            group=group,
            name=field_name,
            added=added,
            removed=removed,
            changed=changed,
            unchanged=unchanged,
        )

    @classmethod
    def _add_missing_hash(cls, client: CogniteClient, resource: list[CDFResource]):
        """
        This function adds the hash to the metadata of the resource if it is missing.
        The intention is for backwards compatibility with resources that were created before the hash was added.
        """
        for item in resource:
            resource = item.cdf_resource
            if not resource.metadata:
                resource.metadata = {}
            if CDFResource.content_key_hash not in resource.metadata:
                if isinstance(item, CDFSequence):
                    content = client.sequences.data.retrieve_dataframe(
                        start=0, end=-1, external_id=item.external_id, limit=None
                    )
                    resource.metadata[CDFResource.content_key_hash] = CDFSequence.calculate_hash(content)
                    update = SequenceUpdate(external_id=item.external_id).metadata.add(
                        {CDFSequence.content_key_hash: resource.metadata[CDFResource.content_key_hash]}
                    )
                    client.sequences.update(update)
                elif isinstance(item, CDFFile):
                    content = client.files.download_bytes(item.external_id)
                    resource.metadata[CDFResource.content_key_hash] = CDFFile.calculate_hash(content)
                    update = FileMetadataUpdate(external_id=item.external_id).metadata.add(
                        {CDFFile.content_key_hash: resource.metadata[CDFResource.content_key_hash]}
                    )
                    client.files.update(update)
                else:
                    raise NotImplementedError(f"Cannot get content for {item}")


T_Model = TypeVar("T_Model", bound=Model)



from __future__ import annotations

from typing import cast, Type

from cognite.powerops.resync.config.resync_config import ReSyncConfig
from cognite.powerops.resync.models.base import AssetModel, DataModel, Model
from cognite.powerops.resync import models
from .to_cogshop_model import to_cogshop_asset_model, to_cogshop_data_model
from .to_market_model import to_benchmark_data_model, to_dayahead_data_model, to_market_asset_model, to_rkom_data_model
from .to_production_model import to_production_data_model, to_production_model


def transform(
    config: ReSyncConfig,
    market_name: str,
    model_types: set[Type[Model]],
) -> list[Model]:
    asset_models: list[AssetModel] = []
    data_models: list[DataModel] = []

    # The Production model is a prerequisite for the Market and CogShop models
    has_asset_model = any(issubclass(m, (AssetModel, models.CogShop1Asset)) for m in model_types)
    if has_asset_model:
        production_model = to_production_model(config.production)
        if models.ProductionModel in model_types:
            asset_models.append(production_model)
        if models.MarketModel in model_types:
            market_model = to_market_asset_model(config.market, production_model.price_areas, market_name)
            settings = config.settings
            market_model.set_root_asset(
                settings.shop_service_url,
                settings.organization_subdomain,
                settings.tenant_id,
                production_model.root_asset.external_id,
            )
            asset_models.append(market_model)
        if models.CogShop1Asset in model_types:
            cogshop_model = to_cogshop_asset_model(
                config.cogshop, production_model.watercourses, config.settings.shop_version
            )
            data_models.append(cogshop_model)

    has_data_model = any(issubclass(m, DataModel) and not issubclass(m, models.CogShop1Asset) for m in model_types)
    if has_data_model:
        # The production model is a prerequisite for the CogShop and Market models
        production__data_model = to_production_data_model(config.production)
        if models.ProductionModelDM in model_types:
            data_models.append(production__data_model)
        if models.CogShopDataModel in model_types:
            cogshop_data_model = to_cogshop_data_model(
                config.cogshop, config.production.watercourses, config.settings.shop_version
            )
            data_models.append(cogshop_data_model)
        if models.BenchmarkMarketDataModel in model_types or models.DayAheadMarketDataModel in model_types:
            benchmark_market_model = to_benchmark_data_model(config.market.benchmarks)
            if models.BenchmarkMarketDataModel in model_types:
                data_models.append(benchmark_market_model)
            if models.DayAheadMarketDataModel in model_types:
                dayahead_benchmark = benchmark_market_model.benchmarking[0]
                dayahead_bid = benchmark_market_model.bids[cast(str, dayahead_benchmark.bid)]
                day_ahead_market_model = to_dayahead_data_model(
                    config.market, dayahead_benchmark, dayahead_bid, production__data_model.price_areas
                )
                data_models.append(day_ahead_market_model)
        if models.RKOMMarketDataModel in model_types:
            rkom_market_model = to_rkom_data_model(config.market, market_name)
            data_models.append(rkom_market_model)

    all_models: list[Model] = cast(list[Model], asset_models) + cast(list[Model], data_models)
    return all_models



from __future__ import annotations

from collections.abc import Sequence
from typing import Generic, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.cogshop1.data_classes._core import T_TypeApplyNode, T_TypeNode, T_TypeNodeList


class TypeAPI(Generic[T_TypeNode, T_TypeApplyNode, T_TypeNodeList]):
    def __init__(
        self,
        client: CogniteClient,
        sources: dm.ViewIdentifier | Sequence[dm.ViewIdentifier] | dm.View | Sequence[dm.View],
        class_type: type[T_TypeNode],
        class_apply_type: type[T_TypeApplyNode],
        class_list: type[T_TypeNodeList],
    ):
        self._client = client
        self.sources = sources
        self.class_type = class_type
        self.class_apply_type = class_apply_type
        self.class_list = class_list

    @overload
    def _retrieve(self, external_id: str) -> T_TypeNode:
        ...

    @overload
    def _retrieve(self, external_id: Sequence[str]) -> T_TypeNodeList:
        ...

    def _retrieve(
        self, nodes: dm.NodeId | Sequence[dm.NodeId] | tuple[str, str] | Sequence[tuple[str, str]]
    ) -> T_TypeNode | T_TypeNodeList:
        is_multiple = (
            isinstance(nodes, Sequence)
            and not isinstance(nodes, str)
            and not (isinstance(nodes, tuple) and isinstance(nodes[0], str))
        )
        instances = self._client.data_modeling.instances.retrieve(nodes=nodes, sources=self.sources)
        if is_multiple:
            return self.class_list([self.class_type.from_node(node) for node in instances.nodes])
        return self.class_type.from_node(instances.nodes[0])

    def _list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT, filter: dm.Filter | None = None) -> T_TypeNodeList:
        nodes = self._client.data_modeling.instances.list("node", sources=self.sources, limit=limit, filter=filter)
        return self.class_list([self.class_type.from_node(node) for node in nodes])



from __future__ import annotations

import itertools
from dataclasses import dataclass, field, asdict
from itertools import islice
from typing import Literal

import pandas as pd
from cognite.client.data_classes import Sequence, FileMetadata
from cognite.client.data_classes._base import CogniteResource
from deepdiff import DeepDiff  # type: ignore[import]

from cognite.powerops.resync.models.base.resource_type import Resource
from cognite.powerops.resync.models.cdf_resources import CDFResource


@dataclass
class Change:
    last: Resource
    new: Resource

    @property
    def changed_fields(self) -> str:
        last_dumped = (
            self.last.dump(camel_case=True) if isinstance(self.last, CogniteResource) else self.last.model_dump()
        )
        new_dumped = self.new.dump(camel_case=True) if isinstance(self.new, CogniteResource) else self.new.model_dump()
        return (
            DeepDiff(last_dumped, new_dumped, ignore_string_case=True)
            .pretty()
            .replace("added to dictionary.", f"will be added to {self.last.external_id}.")
            .replace("removed from dictionary.", f"will be removed from {self.last.external_id}.")
        )

    @property
    def is_changed_content(self) -> bool:
        if isinstance(self.last, (Sequence, FileMetadata)) and isinstance(self.new, (Sequence, FileMetadata)):
            last_hash = self.last.metadata.get(CDFResource.content_key_hash) if self.last.metadata else None
            new_hash = self.new.metadata.get(CDFResource.content_key_hash) if self.new.metadata else None
            return last_hash != new_hash or last_hash is None or new_hash is None
        return False


@dataclass
class FieldSummary:
    group: Literal["CDF", "Domain"]
    name: str
    added: int
    removed: int
    changed: int
    unchanged: int

    @property
    def total(self) -> int:
        return self.added + self.removed + self.changed + self.unchanged


@dataclass
class FieldIds:
    group: Literal["CDF", "Domain"]
    name: str
    added: list[str] = field(default_factory=list)
    removed: list[str] = field(default_factory=list)
    changed: list[str] = field(default_factory=list)
    unchanged: list[str] = field(default_factory=list)

    @property
    def total(self) -> int:
        return len(self.added) + len(self.removed) + len(self.changed) + len(self.unchanged)


@dataclass
class FieldDifference:
    group: Literal["CDF", "Domain"]
    name: str
    added: list[Resource] = field(default_factory=list)
    removed: list[Resource] = field(default_factory=list)
    changed: list[Change] = field(default_factory=list)
    unchanged: list[Resource] = field(default_factory=list)

    @property
    def total(self) -> int:
        return len(self.added) + len(self.removed) + len(self.changed) + len(self.unchanged)

    def as_summary(self) -> FieldSummary:
        return FieldSummary(
            group=self.group,
            name=self.name,
            added=len(self.added),
            removed=len(self.removed),
            changed=len(self.changed),
            unchanged=len(self.unchanged),
        )

    def as_ids(self, limit: int = -1) -> FieldIds:
        def get_identifier(item: Resource) -> str:
            if hasattr(item, "external_id"):
                return item.external_id or "missing external_id"
            elif hasattr(item, "name"):
                return item.name or "missing name"
            raise NotImplementedError(f"Could not find external_id or name in {item}")

        limit_ = limit if limit > 0 else None
        return FieldIds(
            group=self.group,
            name=self.name,
            added=[get_identifier(item) for item in islice(self.added, limit_)],
            removed=[get_identifier(item) for item in islice(self.removed, limit_)],
            changed=[get_identifier(item.last) for item in islice(self.changed, limit_)],
            unchanged=[get_identifier(item) for item in islice(self.unchanged, limit_)],
        )

    def set_set_dataset(self, dataset: int) -> None:
        for item in itertools.chain(self.added, self.removed, self.unchanged):
            if hasattr(item, "data_set_id"):
                item.data_set_id = dataset
        for item in self.changed:
            if hasattr(item.last, "data_set_id"):
                item.last.data_set_id = dataset
                item.new.data_set_id = dataset


@dataclass
class ModelDifference:
    name: str
    changes: list[FieldDifference]

    def has_changes(self, exclude: set | frozenset = frozenset({"timeseries"})) -> bool:
        return any(change.total != change.unchanged for change in self.changes if change.name not in exclude)


@dataclass
class ModelDifferences:
    models: list[ModelDifference]

    def has_changes(self, exclude: set | frozenset = frozenset({"timeseries"})) -> bool:
        return any(m.has_changes(exclude) for m in self.models)

    def as_markdown(self) -> str:
        report = []
        for model in self.models:
            summaries = [change.as_summary() for change in model.changes]
            table = pd.DataFrame(
                [
                    _exclude_keys(asdict(summary), keys=["group", "name"])
                    for summary in summaries
                    if summary.group == "CDF"
                ],
                index=[summary.name for summary in summaries if summary.group == "CDF"],
            )
            ids = [change.as_ids(limit=5) for change in model.changes if change.group == "CDF"]
            added_ids = []
            for id_ in ids:
                if id_.added:
                    added_ids.append(f"#### Added {id_.name}\n{id_.added}")
            removed_ids = []
            for id_ in ids:
                if id_.removed:
                    removed_ids.append(f"#### Removed {id_.name}\n{id_.removed}")
            changed_samples = []
            for change, id_ in zip((c for c in model.changes if c.group == "CDF"), ids):
                if change.changed:
                    new_line = "\n * "
                    sample_line = new_line.join(
                        f"{i}: {c.changed_fields}" for c, i in zip(change.changed[:3], id_.changed[:3])
                    )
                    changed_samples.append(f"#### Changed {change.name}\n{sample_line}")
            samples = []
            if added_ids:
                samples.append("\n".join(added_ids))
            if removed_ids:
                samples.append("\n".join(removed_ids))
            if changed_samples:
                samples.append("\n".join(changed_samples))
            if samples:
                samples.insert(0, "### Samples")
            samples = "\n".join(samples)

            report.append(
                f"""## {model.name}

{table.T.to_markdown()}

    **NOTE** Timeseries are not updated by `resync`

{samples}
    """
            )
        return "\n".join(report)


def _exclude_keys(d: dict, keys: list[str]) -> dict:
    return {k: v for k, v in d.items() if k not in keys}



from __future__ import annotations

import json
import logging
import os
from datetime import datetime, timezone
from typing import BinaryIO, Literal, TextIO, Union

import requests
from cognite.client import CogniteClient
from cognite.client.data_classes import Event, FileMetadata
from cognite.client.utils._time import time_ago_to_ms, datetime_to_ms

from cognite.powerops.cdf_labels import RelationshipLabel
from cognite.powerops.clients.data_set_api import DataSetsAPI
from cognite.powerops.clients.shop.api.shop_results_api import ShopRunResultsAPI
from cognite.powerops.clients.shop.data_classes import Case, ShopRun, ShopRunEvent
from cognite.powerops.utils.cdf.calls import retrieve_event
from cognite.powerops.utils.cdf.resource_creation import simple_relationship

logger = logging.getLogger(__name__)


InputFileTypeT = Literal["case", "cut", "mapping", "extra"]


RUN_SHOP_URL = "https://power-ops-api.staging.{cluster}.cognite.ai/{project}/run-shop"


class ShopRunsAPI:
    def __init__(
        self, client: CogniteClient, data_set_api: DataSetsAPI, result_api: ShopRunResultsAPI, cogshop_version: str
    ):
        self._client = client
        self._data_set_api = data_set_api
        self._result_api = result_api
        self._cogshop_version = cogshop_version

    def trigger(self, case: Case) -> ShopRun:
        logger.info("Triggering SHOP run...")
        shop_run = self._upload_to_cdf(case)
        self._post_shop_run(shop_run.shop_run_event.external_id)
        return shop_run

    def _upload_to_cdf(self, case: Case) -> ShopRun:
        shop_run_event = ShopRunEvent(
            watercourse="",
            starttime=case.data["time"]["starttime"],
            endtime=case.data["time"]["endtime"],
            timeresolution=case.data["time"]["timeresolution"],
            manual_run=True,
        )
        logger.debug(f"Uploading event '{shop_run_event.external_id}'.")
        case_file_meta = self._upload_input_file_bytes(case.yaml.encode(), shop_run_event, file_type="case")
        event = shop_run_event.to_event(self._data_set_api.write_dataset_id)
        self._connect_file_to_event(case_file_meta, event, RelationshipLabel.CASE_FILE.value)
        preprocessor_metadata = {"cog_shop_case_file": {"external_id": case_file_meta.external_id}}

        if case.cut_file:
            cut_file_meta = self._upload_input_file(
                case.cut_file["file"], shop_run_event=shop_run_event, input_file_type="cut"
            )
            self._connect_file_to_event(cut_file_meta, event, RelationshipLabel.CUT_FILE.value)
            preprocessor_metadata["cut_file"] = {"external_id": cut_file_meta.external_id}

        if case.mapping_files:
            preprocessor_metadata["mapping_files"] = []
        for mapping_file in case.mapping_files or []:
            mapping_file_meta = self._upload_input_file(
                mapping_file["file"], shop_run_event=shop_run_event, input_file_type="mapping"
            )
            self._connect_file_to_event(mapping_file_meta, event, RelationshipLabel.MAPPING_FILE.value)
            preprocessor_metadata["mapping_files"].append({"external_id": mapping_file_meta.external_id})

        if case.extra_files:
            preprocessor_metadata["extra_files"] = []
        for extra_file in case.extra_files or []:
            extra_file_meta = self._upload_input_file(
                extra_file["file"], shop_run_event=shop_run_event, input_file_type="extra"
            )
            self._connect_file_to_event(extra_file_meta, event, RelationshipLabel.EXTRA_FILE.value)
            preprocessor_metadata["extra_files"].append({"external_id": extra_file_meta.external_id})

        event.metadata["shop:preprocessor_data"] = json.dumps(preprocessor_metadata)
        event.metadata["processed"] = "yes"  # avoid event being picked up by sniffer
        self._client.events.create(event)
        return ShopRun(self.retrieve_status, retrieve_results=self._result_api.retrieve, shop_run_event=shop_run_event)

    def _upload_input_file(
        self, file: str, shop_run_event: ShopRunEvent, input_file_type: InputFileTypeT
    ) -> FileMetadata:
        _, input_file_ext = os.path.splitext(file)
        with open(file, "rb") as file_stream:
            return self._upload_input_file_bytes(file_stream.read(), shop_run_event, input_file_type, input_file_ext)

    def _upload_input_file_bytes(
        self,
        content: Union[str, bytes, TextIO, BinaryIO],
        shop_run_event: ShopRunEvent,
        file_type: InputFileTypeT,
        file_ext: str = ".yaml",
    ) -> FileMetadata:
        file_ext_id = f"{shop_run_event.external_id}_{file_type.upper()}"
        logger.debug(f"Uploading file: '{file_ext_id}'.")
        return self._client.files.upload_bytes(
            external_id=file_ext_id,
            content=content,
            data_set_id=self._data_set_api.write_dataset_id,
            name=f"{shop_run_event.external_id}_{file_type}{file_ext}",
            mime_type="application/yaml",
            metadata={
                "shop:run_event_id": shop_run_event.external_id,
                "shop:type": f"cog_shop_{file_type}",
            },
            overwrite=True,
        )

    def _connect_file_to_event(self, file_meta: FileMetadata, event: Event, relationship_label: str):
        relationship = simple_relationship(
            source=event,
            target=file_meta,
            label_external_id=relationship_label,
        )
        relationship.data_set_id = self._data_set_api.write_dataset_id
        self._client.relationships.create(relationship)

    def _post_shop_run(self, shop_run_external_id: str):
        logger.info(f"Triggering run-shop endpoint, cogShopVersion: '{self._cogshop_version}'.")
        cdf_config = self._client.config
        project = cdf_config.project
        cluster = cdf_config.base_url.split("//")[1].split(".")[0]
        auth_header = dict([cdf_config.credentials.authorization_header()])

        response = requests.post(
            RUN_SHOP_URL.format(cluster=cluster, project=project),
            json={
                "shopEventExternalId": shop_run_external_id,
                "datasetId": self._data_set_api.write_dataset_id,
                "cogShopVersion": self._cogshop_version,
            },
            headers=auth_header,
        )
        response.raise_for_status()
        logger.debug(response.json())

    def list(self, created_after: int | str | None = None, limit: int | None = 25) -> list[ShopRun]:
        if isinstance(created_after, str):
            timespan = time_ago_to_ms(created_after)
            now = datetime_to_ms(datetime.now(timezone.utc))
            created_after = now - timespan

        shop_run_events = self._client.events.list(
            type=ShopRunEvent.event_type,
            created_time={"min": created_after} if created_after is not None else None,
            metadata={"process_type": ShopRunEvent.process_type},
            limit=limit,
        )

        return [
            ShopRun(self.retrieve_status, retrieve_results=self._result_api.retrieve, shop_run_event=shop_run_event)
            for shop_run_event in shop_run_events
        ]

    def retrieve(self, external_id: str) -> ShopRun:
        event = self._client.events.retrieve(external_id=external_id)
        return ShopRun(
            self.retrieve_status,
            retrieve_results=self._result_api.retrieve,
            shop_run_event=ShopRunEvent.from_event(event),
        )

    def retrieve_status(self, shop_run_external_id: str) -> ShopRun.Status:
        event = retrieve_event(self._client, shop_run_external_id)
        logger.debug(f"Reading status from event {event.external_id}.")
        if relationships := self._client.relationships.list(
            # At the moment, looking at run whose related event might not be in `uc:000:powerops`
            # data_set_ids=[self._data_set_api.read_dataset_id],
            source_external_ids=[shop_run_external_id],
            target_types=["event"],
        ):
            related_events = self._client.events.retrieve_multiple(
                external_ids=[rel.target_external_id for rel in relationships],
                ignore_unknown_ids=True,
            )
        else:
            related_events = []
        if not len(related_events) or all(ev.type == "POWEROPS_PROCESS_STARTED" for ev in related_events):
            return ShopRun.Status.IN_PROGRESS
        elif any(ev.type == "POWEROPS_PROCESS_FINISHED" for ev in related_events):
            return ShopRun.Status.SUCCEEDED
        else:
            return ShopRun.Status.FAILED



from __future__ import annotations

from typing import ClassVar, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

__all__ = ["Generator", "GeneratorApply", "GeneratorList"]


class Generator(DomainModel):
    space: ClassVar[str] = "power-ops"
    generator_efficiency_curve: Optional[str] = Field(None, alias="generatorEfficiencyCurve")
    name: Optional[str] = None
    p_min: Optional[float] = Field(None, alias="pMin")
    penstock: Optional[int] = None
    start_stop_cost: Optional[str] = Field(None, alias="startStopCost")
    is_available: Optional[str] = Field(None, alias="isAvailable")
    startcost: Optional[float] = None
    turbine_efficiency_curve: Optional[str] = Field(None, alias="turbineEfficiencyCurve")


class GeneratorApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    generator_efficiency_curve: Optional[str] = None
    name: Optional[str] = None
    p_min: Optional[float] = None
    penstock: Optional[int] = None
    start_stop_cost: Optional[str] = None
    is_available: Optional[str] = None
    startcost: Optional[float] = None
    turbine_efficiency_curve: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "Generator"),
            properties={
                "generatorEfficiencyCurve": self.generator_efficiency_curve,
                "name": self.name,
                "pMin": self.p_min,
                "penstock": self.penstock,
                "startStopCost": self.start_stop_cost,
                "startcost": self.startcost,
                "turbineEfficiencyCurve": self.turbine_efficiency_curve,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class GeneratorList(TypeList[Generator]):
    _NODE = Generator



from __future__ import annotations

from cognite.powerops.resync.config.market._core import PriceScenario, PriceScenarioID
from cognite.powerops.resync.config.market.benchmarking import BenchmarkingConfig



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.clients.data_classes._input_time_series_mappings import InputTimeSeriesMappingApply

__all__ = ["ScenarioMapping", "ScenarioMappingApply", "ScenarioMappingList"]


class ScenarioMapping(DomainModel):
    space: ClassVar[str] = "power-ops"
    mapping_override: list[str] = Field([], alias="mappingOverride")
    name: Optional[str] = None
    shop_type: Optional[str] = Field(None, alias="shopType")
    watercourse: Optional[str] = None


class ScenarioMappingApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    mapping_override: list[Union["InputTimeSeriesMappingApply", str]] = Field(default_factory=list, repr=False)
    name: Optional[str] = None
    shop_type: Optional[str] = None
    watercourse: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "ScenarioMapping"),
            properties={
                "name": self.name,
                "shopType": self.shop_type,
                "watercourse": self.watercourse,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        for mapping_override in self.mapping_override:
            edge = self._create_mapping_override_edge(mapping_override)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(mapping_override, DomainModelApply):
                instances = mapping_override._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_mapping_override_edge(
        self, mapping_override: Union[str, "InputTimeSeriesMappingApply"]
    ) -> dm.EdgeApply:
        if isinstance(mapping_override, str):
            end_node_ext_id = mapping_override
        elif isinstance(mapping_override, DomainModelApply):
            end_node_ext_id = mapping_override.external_id
        else:
            raise TypeError(f"Expected str or InputTimeSeriesMappingApply, got {type(mapping_override)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "ScenarioMapping.mappingOverride"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class ScenarioMappingList(TypeList[ScenarioMapping]):
    _NODE = ScenarioMapping



from __future__ import annotations

import ast
import json
import typing
from typing import ClassVar, Dict, List, Optional

from pydantic import BaseModel, Field, field_validator

from cognite.powerops.resync.config.market import PriceScenarioID
from cognite.powerops.resync.config.market._core import Configuration, RelativeTime


class BidMatrixGeneratorConfig(BaseModel):
    name: str
    default_method: str
    default_function_external_id: str
    column_external_ids: ClassVar[list[str]] = [
        "shop_plant",
        "bid_matrix_generation_method",
        "function_external_id",
    ]


class BidProcessConfig(Configuration):
    name: str
    price_area_name: str = Field(alias="bid_price_area")
    price_scenarios: List[PriceScenarioID] = Field(alias="bid_price_scenarios")
    main_scenario: str = Field(alias="bid_main_scenario")
    bid_date: Optional[RelativeTime] = None
    shop_start: Optional[RelativeTime] = Field(None, alias="shop_starttime")
    shop_end: Optional[RelativeTime] = Field(None, alias="shop_endtime")
    bid_matrix_generator: str = Field(alias="bid_bid_matrix_generator_config_external_id")
    price_scenarios_per_watercourse: Optional[Dict[str, typing.Set[str]]] = None
    is_default_config_for_price_area: bool = False
    no_shop: bool = Field(False, alias="no_shop")

    @field_validator("shop_start", "shop_end", "bid_date", mode="before")
    def json_loads(cls, value):
        return {"operations": json.loads(value)} if isinstance(value, str) else value

    @field_validator("price_scenarios", mode="before")
    def literal_eval(cls, value):
        return [{"id": id_} for id_ in ast.literal_eval(value)] if isinstance(value, str) else value



from typing import ClassVar, Optional


from cognite.client.data_classes import Asset
from pydantic import field_validator
from pydantic.dataclasses import Field

from cognite.powerops.resync.models.base import AssetModel

from .base import Bid, DateTransformation, Market, Process, ShopTransformation
from .benchmark import BenchmarkBid, BenchmarkProcess, ProductionPlanTimeSeries
from .dayahead import DayAheadBid, DayAheadProcess, NordPoolMarket
from .rkom import RKOMBid, RKOMBidCombination, RKOMCombinationBid, RKOMMarket, RKOMPlants, RKOMProcess
from ..base.asset_type import T_Asset_Type


class MarketModel(AssetModel):
    root_asset: ClassVar[Optional[Asset]] = None
    markets: list[Market] = Field(default_factory=list)
    dayahead_processes: list[DayAheadProcess] = Field(default_factory=list)
    benchmark_processes: list[BenchmarkProcess] = Field(default_factory=list)
    rkom_processes: list[RKOMProcess] = Field(default_factory=list)
    combinations: list[RKOMBidCombination] = Field(default_factory=list)

    @field_validator(
        "markets", "dayahead_processes", "benchmark_processes", "rkom_processes", "combinations", mode="after"
    )
    def ordering(cls, value: list[T_Asset_Type]) -> list[T_Asset_Type]:
        # To ensure loading the production model always yields the same result, we sort the assets by external_id.
        return sorted(value, key=lambda x: x.external_id)

    @classmethod
    def set_root_asset(
        cls, shop_service_url: str, organization_subdomain: str, tenant_id: str, core_root_asset_external_id: str
    ) -> None:
        if shop_service_url == "https://shop-staging.az-inso-powerops.cognite.ai/submit-run":
            customer = "cognite"
        else:
            customer = organization_subdomain

        cls.root_asset = Asset(
            parent_external_id=core_root_asset_external_id,
            external_id="configurations",
            name="Configurations",
            description="Configurations used for PowerOps",
            metadata={
                "shop_service_url": shop_service_url,
                "organization_subdomain": organization_subdomain,
                "customer": customer,
                "tenant_id": tenant_id,
            },
        )

    def standardize(self) -> None:
        self.markets = self.ordering(self.markets)
        self.dayahead_processes = self.ordering(self.dayahead_processes)
        self.benchmark_processes = self.ordering(self.benchmark_processes)
        self.rkom_processes = self.ordering(self.rkom_processes)
        self.combinations = self.ordering(self.combinations)
        for field in [
            self.markets,
            self.dayahead_processes,
            self.benchmark_processes,
            self.rkom_processes,
            self.combinations,
        ]:
            for item in field:
                item.standardize()


__all__ = [
    "RKOMBid",
    "RKOMBidCombination",
    "RKOMMarket",
    "RKOMProcess",
    "RKOMCombinationBid",
    "RKOMPlants",
    "DayAheadBid",
    "NordPoolMarket",
    "DayAheadProcess",
    "BenchmarkBid",
    "BenchmarkProcess",
    "ProductionPlanTimeSeries",
    "Market",
    "DateTransformation",
    "ShopTransformation",
    "Bid",
    "Process",
    "MarketModel",
]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.cogshop1.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.cogshop1.data_classes._processing_logs import ProcessingLogApply
    from cognite.powerops.cogshop1.data_classes._scenarios import ScenarioApply

__all__ = ["Case", "CaseApply", "CaseList"]


class Case(DomainModel):
    space: ClassVar[str] = "cogShop"
    end_time: Optional[str] = Field(None, alias="endTime")
    processing_log: list[str] = Field([], alias="processingLog")
    scenario: Optional[str] = None
    start_time: Optional[str] = Field(None, alias="startTime")


class CaseApply(DomainModelApply):
    space: ClassVar[str] = "cogShop"
    end_time: str
    processing_log: list[Union["ProcessingLogApply", str]] = Field(default_factory=list, repr=False)
    scenario: Optional[Union["ScenarioApply", str]] = Field(None, repr=False)
    start_time: str

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("cogShop", "Case"),
            properties={
                "endTime": self.end_time,
                "scenario": {
                    "space": "cogShop",
                    "externalId": self.scenario if isinstance(self.scenario, str) else self.scenario.external_id,
                },
                "startTime": self.start_time,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        for processing_log in self.processing_log:
            edge = self._create_processing_log_edge(processing_log)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(processing_log, DomainModelApply):
                instances = processing_log._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.scenario, DomainModelApply):
            instances = self.scenario._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_processing_log_edge(self, processing_log: Union[str, "ProcessingLogApply"]) -> dm.EdgeApply:
        if isinstance(processing_log, str):
            end_node_ext_id = processing_log
        elif isinstance(processing_log, DomainModelApply):
            end_node_ext_id = processing_log.external_id
        else:
            raise TypeError(f"Expected str or ProcessingLogApply, got {type(processing_log)}")

        return dm.EdgeApply(
            space="cogShop",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("cogShop", "Case.processingLog"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("cogShop", end_node_ext_id),
        )


class CaseList(TypeList[Case]):
    _NODE = Case



from __future__ import annotations

from pathlib import Path

from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import OAuthClientCredentials

from cognite.powerops.clients._api.benchmark_bids import BenchmarkBidsAPI
from cognite.powerops.clients._api.benchmark_process import BenchmarkProcessAPI
from cognite.powerops.clients._api.bid_matrix_generators import BidMatrixGeneratorsAPI
from cognite.powerops.clients._api.command_configs import CommandConfigsAPI
from cognite.powerops.clients._api.date_transformations import DateTransformationsAPI
from cognite.powerops.clients._api.day_ahead_bids import DayAheadBidsAPI
from cognite.powerops.clients._api.day_ahead_process import DayAheadProcessAPI
from cognite.powerops.clients._api.generators import GeneratorsAPI
from cognite.powerops.clients._api.input_time_series_mappings import InputTimeSeriesMappingsAPI
from cognite.powerops.clients._api.nord_pool_markets import NordPoolMarketsAPI
from cognite.powerops.clients._api.output_containers import OutputContainersAPI
from cognite.powerops.clients._api.output_mappings import OutputMappingsAPI
from cognite.powerops.clients._api.plants import PlantsAPI
from cognite.powerops.clients._api.price_areas import PriceAreasAPI
from cognite.powerops.clients._api.production_plan_time_series import ProductionPlanTimeSeriesAPI
from cognite.powerops.clients._api.rkom_bids import RKOMBidsAPI
from cognite.powerops.clients._api.rkom_bid_combinations import RKOMBidCombinationsAPI
from cognite.powerops.clients._api.rkom_combination_bids import RKOMCombinationBidsAPI
from cognite.powerops.clients._api.rkom_markets import RKOMMarketsAPI
from cognite.powerops.clients._api.rkom_process import RKOMProcessAPI
from cognite.powerops.clients._api.reserve_scenarios import ReserveScenariosAPI
from cognite.powerops.clients._api.reservoirs import ReservoirsAPI
from cognite.powerops.clients._api.scenarios import ScenariosAPI
from cognite.powerops.clients._api.scenario_mappings import ScenarioMappingsAPI
from cognite.powerops.clients._api.scenario_templates import ScenarioTemplatesAPI
from cognite.powerops.clients._api.shop_transformations import ShopTransformationsAPI
from cognite.powerops.clients._api.value_transformations import ValueTransformationsAPI
from cognite.powerops.clients._api.watercourses import WatercoursesAPI
from cognite.powerops.clients._api.watercourse_shops import WatercourseShopsAPI


class BenchmarkAPIs:
    """
    BenchmarkAPIs

    Data Model:
        space: power-ops
        externalId: benchmarkMarket
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.benchmark_bids = BenchmarkBidsAPI(client)
        self.benchmark_process = BenchmarkProcessAPI(client)
        self.date_transformations = DateTransformationsAPI(client)
        self.input_time_series_mappings = InputTimeSeriesMappingsAPI(client)
        self.nord_pool_markets = NordPoolMarketsAPI(client)
        self.production_plan_time_series = ProductionPlanTimeSeriesAPI(client)
        self.scenario_mappings = ScenarioMappingsAPI(client)
        self.shop_transformations = ShopTransformationsAPI(client)
        self.value_transformations = ValueTransformationsAPI(client)


class CogShopAPIs:
    """
    CogShopAPIs

    Data Model:
        space: power-ops
        externalId: cogshop
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.command_configs = CommandConfigsAPI(client)
        self.input_time_series_mappings = InputTimeSeriesMappingsAPI(client)
        self.output_containers = OutputContainersAPI(client)
        self.output_mappings = OutputMappingsAPI(client)
        self.scenarios = ScenariosAPI(client)
        self.scenario_mappings = ScenarioMappingsAPI(client)
        self.scenario_templates = ScenarioTemplatesAPI(client)
        self.value_transformations = ValueTransformationsAPI(client)


class DayAheadAPIs:
    """
    DayAheadAPIs

    Data Model:
        space: power-ops
        externalId: dayaheadMarket
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.bid_matrix_generators = BidMatrixGeneratorsAPI(client)
        self.date_transformations = DateTransformationsAPI(client)
        self.day_ahead_bids = DayAheadBidsAPI(client)
        self.day_ahead_process = DayAheadProcessAPI(client)
        self.input_time_series_mappings = InputTimeSeriesMappingsAPI(client)
        self.nord_pool_markets = NordPoolMarketsAPI(client)
        self.scenario_mappings = ScenarioMappingsAPI(client)
        self.shop_transformations = ShopTransformationsAPI(client)
        self.value_transformations = ValueTransformationsAPI(client)


class ProductionAPIs:
    """
    ProductionAPIs

    Data Model:
        space: power-ops
        externalId: production
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.generators = GeneratorsAPI(client)
        self.plants = PlantsAPI(client)
        self.price_areas = PriceAreasAPI(client)
        self.reservoirs = ReservoirsAPI(client)
        self.watercourses = WatercoursesAPI(client)
        self.watercourse_shops = WatercourseShopsAPI(client)


class RKOMMarketAPIs:
    """
    RKOMMarketAPIs

    Data Model:
        space: power-ops
        externalId: rkomMarket
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.date_transformations = DateTransformationsAPI(client)
        self.input_time_series_mappings = InputTimeSeriesMappingsAPI(client)
        self.rkom_bids = RKOMBidsAPI(client)
        self.rkom_bid_combinations = RKOMBidCombinationsAPI(client)
        self.rkom_combination_bids = RKOMCombinationBidsAPI(client)
        self.rkom_markets = RKOMMarketsAPI(client)
        self.rkom_process = RKOMProcessAPI(client)
        self.reserve_scenarios = ReserveScenariosAPI(client)
        self.scenario_mappings = ScenarioMappingsAPI(client)
        self.shop_transformations = ShopTransformationsAPI(client)
        self.value_transformations = ValueTransformationsAPI(client)


class PowerOpsClient:
    """
    PowerOpsClient

    Generated with:
        pygen = 0.15.2
        cognite-sdk = 6.13.3
        pydantic = 2.1.1

    """

    def __init__(self, config_or_client: CogniteClient | ClientConfig):
        if isinstance(config_or_client, CogniteClient):
            client = config_or_client
        elif isinstance(config_or_client, ClientConfig):
            client = CogniteClient(config_or_client)
        else:
            raise ValueError(f"Expected CogniteClient or ClientConfig, got {type(config_or_client)}")
        self.benchmark = BenchmarkAPIs(client)
        self.cog_shop = CogShopAPIs(client)
        self.day_ahead = DayAheadAPIs(client)
        self.production = ProductionAPIs(client)
        self.rkom_market = RKOMMarketAPIs(client)

    @classmethod
    def azure_project(
        cls, tenant_id: str, client_id: str, client_secret: str, cdf_cluster: str, project: str
    ) -> PowerOpsClient:
        credentials = OAuthClientCredentials.default_for_azure_ad(tenant_id, client_id, client_secret, cdf_cluster)
        config = ClientConfig.default(project, cdf_cluster, credentials)

        return cls(config)

    @classmethod
    def from_toml(cls, file_path: Path | str, section: str | None = "cognite") -> PowerOpsClient:
        import toml

        toml_content = toml.load(file_path)
        if section is not None:
            try:
                toml_content = toml_content[section]
            except KeyError:
                raise ValueError(f"Could not find section '{section}' in {file_path}")

        return cls.azure_project(**toml_content)



from __future__ import annotations

import logging
from typing import Optional, Union

from cognite.client import CogniteClient
from cognite.client._api.assets import AssetsAPI
from cognite.client._api.events import EventsAPI
from cognite.client._api.files import FilesAPI
from cognite.client._api.labels import LabelsAPI
from cognite.client._api.relationships import RelationshipsAPI
from cognite.client._api.sequences import SequencesAPI
from cognite.client.data_classes import Asset, Event, LabelDefinition, Relationship, RelationshipList, Sequence
from cognite.client.exceptions import CogniteDuplicatedError

logger = logging.getLogger(__name__)

CogniteAPI = Union[
    AssetsAPI,
    EventsAPI,
    FilesAPI,
    SequencesAPI,
    LabelsAPI,
    RelationshipsAPI,
]


def upsert_cognite_resources(
    api: CogniteAPI,
    resource_type: str,
    resources: list[Asset | Event | Sequence | LabelDefinition | Relationship],
) -> None:
    try:
        api.create(resources)
    except CogniteDuplicatedError as e:
        duplicated_ids = [ids["externalId"] for ids in e.duplicated]

        # Since the non-existing resource are not created when create fails, we need to create them in a separate call.
        if failed_resources := [r for r in e.failed if r.external_id not in duplicated_ids]:
            api.create(failed_resources)

        logger.debug(f"The following {resource_type} already exists in CDF: {duplicated_ids}")
        resources_to_update = [r for r in resources if r.external_id in duplicated_ids]
        if hasattr(api, "update"):
            api.update(resources_to_update)
        elif resource_type == "label_definitions":
            # For labels, the update method is not implemented. Hence, we need to delete and recreate.
            logger.debug("Deleting existing label definitions")
            api.delete(external_id=duplicated_ids)
            logger.debug("Creating new label definitions")
            api.create(resources_to_update)


def retrieve_event(client: CogniteClient, external_id: str) -> Event:
    event = client.events.retrieve(external_id=external_id)
    if event is None:
        raise ValueError(f"Event not found: {external_id}")
    return event


def retrieve_relationships_from_source_ext_id(
    client: CogniteClient,
    source_ext_id: str,
    label_ext_id: Optional[Union[str, list[str]]],
    target_types: Sequence[str] = None,
) -> RelationshipList:
    """
    Retrieve relationships between source and target using the source ext id.
    Using the `containsAny` filter, we can retrieve all relationships with  given label.
    """
    if isinstance(label_ext_id, str):
        label_ext_id = [label_ext_id]
    _labels = None
    if label_ext_id is not None:
        _labels = {"containsAny": [{"externalId": ext_id} for ext_id in label_ext_id]}
    return client.relationships.list(
        source_external_ids=[source_ext_id],
        labels=_labels,
        limit=-1,
        target_types=target_types,
    )



from __future__ import annotations

from pathlib import Path
from typing import ClassVar, Optional, Union, Any

from cognite.client.data_classes import Asset, TimeSeries
from pydantic import ConfigDict, Field, field_validator, field_serializer

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models.base import AssetModel, AssetType, NonAssetType
from cognite.powerops.resync.models.base.asset_type import T_Asset_Type
from cognite.powerops.resync.models.cdf_resources import CDFSequence

from cognite.powerops.resync.utils.serializer import try_load_dict, parse_time_series


class Generator(AssetType):
    parent_external_id: ClassVar[str] = "generators"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.GENERATOR
    p_min: float = 0.0
    penstock: str = ""
    startcost: float = 0.0
    start_stop_cost_time_series: Optional[TimeSeries] = None
    is_available_time_series: Optional[TimeSeries] = None
    generator_efficiency_curve: Optional[CDFSequence] = None
    turbine_efficiency_curve: Optional[CDFSequence] = None

    @field_serializer("start_stop_cost_time_series", "is_available_time_series")
    def ser_time_series(self, value) -> dict[str, Any]:
        if value is None:
            return {}
        return {"externalId": value.external_id}

    @field_validator("generator_efficiency_curve", "turbine_efficiency_curve", mode="before")
    def parse_sequences(cls, value):
        if value == {}:
            return None
        return value

    @field_validator("start_stop_cost_time_series", "is_available_time_series", mode="before")
    def parse_timeseries(cls, value):
        return parse_time_series(value)


class Reservoir(AssetType):
    parent_external_id: ClassVar[str] = "reservoirs"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.RESERVOIR
    display_name: str = ""
    ordering: str = ""


class Plant(AssetType):
    parent_external_id: ClassVar[str] = "plants"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.PLANT
    display_name: str = ""
    ordering: str = ""
    head_loss_factor: float = 0.0
    outlet_level: float = 0.0
    p_min: float = 0.0
    p_max: float = 0.0
    penstock_head_loss_factors: dict = Field(default_factory=dict)
    generators: list[Generator] = Field(default_factory=list)
    inlet_reservoir: Optional[Reservoir] = None
    p_min_time_series: Optional[TimeSeries] = None
    p_max_time_series: Optional[TimeSeries] = None
    water_value_time_series: Optional[TimeSeries] = None
    feeding_fee_time_series: Optional[TimeSeries] = None
    outlet_level_time_series: Optional[TimeSeries] = None
    inlet_level_time_series: Optional[TimeSeries] = None
    head_direct_time_series: Optional[TimeSeries] = None

    @field_validator("generators", mode="after")
    def generator_ordering(cls, value: list[Generator]) -> list[Generator]:
        # To ensure loading the production model always yields the same result, we sort the generators by external_id.
        return sorted(value, key=lambda x: x.external_id)

    def standardize(self) -> None:
        self.generators = self.generator_ordering(self.generators)

    @field_validator("penstock_head_loss_factors", mode="before")
    def parse_str(cls, value) -> dict:
        return try_load_dict(value)

    @field_serializer(
        "p_min_time_series",
        "p_max_time_series",
        "water_value_time_series",
        "feeding_fee_time_series",
        "outlet_level_time_series",
        "inlet_level_time_series",
        "head_direct_time_series",
    )
    def ser_time_series(self, value) -> dict[str, Any]:
        if value is None:
            return {}
        return {"externalId": value.external_id}

    @field_validator(
        "p_min_time_series",
        "p_max_time_series",
        "water_value_time_series",
        "feeding_fee_time_series",
        "outlet_level_time_series",
        "inlet_level_time_series",
        "head_direct_time_series",
        mode="before",
    )
    def parse_timeseries(cls, value):
        return parse_time_series(value)


class WaterCourseShop(NonAssetType):
    penalty_limit: str


class Watercourse(AssetType):
    model_config: ClassVar[ConfigDict] = ConfigDict(protected_namespaces=tuple())
    parent_external_id: ClassVar[str] = "watercourses"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.WATERCOURSE
    shop: WaterCourseShop
    config_version: Optional[str] = Field("", exclude=True)
    model_file: Optional[Path] = Field(None, exclude=True)
    processed_model_file: Optional[Path] = Field(None, exclude=True)
    plants: list[Plant] = Field(default_factory=list)
    production_obligation_time_series: list[TimeSeries] = Field(default_factory=list)

    @field_validator("plants", mode="after")
    def plant_ordering(cls, value: list[Plant]) -> list[Plant]:
        # To ensure loading the production model always yields the same result, we sort the plants by external_id.
        return sorted(value, key=lambda x: x.external_id)

    def standardize(self) -> None:
        self.plants = self.plant_ordering(self.plants)

    @field_validator("production_obligation_time_series", mode="before")
    def none_to_empty_list(cls, value) -> list[TimeSeries]:
        if value is None or (isinstance(value, list) and value and value[0] is None):
            return []
        return value

    @field_serializer("production_obligation_time_series")
    def ser_time_series(self, value) -> dict[str, Any]:
        if value is None:
            return []
        return [{"externalId": ts.external_id} for ts in value]


class PriceArea(AssetType):
    parent_external_id: ClassVar[str] = "price_areas"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.PRICE_AREA
    dayahead_price_time_series: Optional[TimeSeries] = None
    plants: list[Plant] = Field(default_factory=list)
    watercourses: list[Watercourse] = Field(default_factory=list)

    @field_validator("plants", "watercourses", mode="after")
    def ordering(cls, value: list[T_Asset_Type]) -> list[T_Asset_Type]:
        # To ensure loading the production model always yields the same result, we sort the assets by external_id.
        return sorted(value, key=lambda x: x.external_id)

    def standardize(self) -> None:
        self.plants = self.ordering(self.plants)
        self.watercourses = self.ordering(self.watercourses)

    @field_serializer("dayahead_price_time_series")
    def ser_time_series(self, value) -> dict[str, Any]:
        if value is None:
            return {}
        return {"externalId": value.external_id}


class ProductionModel(AssetModel):
    root_asset: ClassVar[Asset] = Asset(external_id="power_ops", name="PowerOps")
    plants: list[Plant] = Field(default_factory=list)
    generators: list[Generator] = Field(default_factory=list)
    reservoirs: list[Reservoir] = Field(default_factory=list)
    watercourses: list[Watercourse] = Field(default_factory=list)
    price_areas: list[PriceArea] = Field(default_factory=list)

    @field_validator("plants", "generators", "reservoirs", "watercourses", "price_areas", mode="after")
    def ordering(cls, value: list[T_Asset_Type]) -> list[T_Asset_Type]:
        # To ensure loading the production model always yields the same result, we sort the assets by external_id.
        return sorted(value, key=lambda x: x.external_id)

    def standardize(self) -> None:
        self.plants = self.ordering(self.plants)
        self.generators = self.ordering(self.generators)
        self.reservoirs = self.ordering(self.reservoirs)
        self.watercourses = self.ordering(self.watercourses)
        self.price_areas = self.ordering(self.price_areas)
        for field in [self.plants, self.generators, self.reservoirs, self.watercourses, self.price_areas]:
            for item in field:
                item.standardize()



from __future__ import annotations

from typing import ClassVar, Optional

from cognite.client import data_modeling as dm

from cognite.powerops.clients.data_classes._core import DomainModel, DomainModelApply, TypeList

__all__ = ["ValueTransformation", "ValueTransformationApply", "ValueTransformationList"]


class ValueTransformation(DomainModel):
    space: ClassVar[str] = "power-ops"
    arguments: Optional[dict] = None
    method: Optional[str] = None


class ValueTransformationApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    arguments: Optional[dict] = None
    method: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "ValueTransformation"),
            properties={
                "arguments": self.arguments,
                "method": self.method,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class ValueTransformationList(TypeList[ValueTransformation]):
    _NODE = ValueTransformation
